import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import (
    DEFAULT_CLEVEL,
    DEFAULT_CODEC,
    DEFAULT_STORE,
    DEFAULT_WORKERS_BACKEND,
)
from dexp.cli.parsing import (
    _get_output_path,
    _parse_channels,
    _parse_devices,
    _parse_slicing,
)
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.deskew import dataset_deskew


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option(
    "--channels",
    "-c",
    default=None,
    help="list of channels for the view in standard order for the microscope type (C0L0, C0L1, C1L0, C1L1,...)",
)
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)  #
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option(
    "--mode", "-m", type=str, default="yang", help="Deskew algorithm: 'yang' or 'classic'. ", show_default=True
)  #
@click.option("--deltax", "-xx", type=float, default=None, help="Pixel size of the camera", show_default=True)  #
@click.option(
    "--deltaz",
    "-zz",
    type=float,
    default=None,
    help="Scanning step (stage or galvo scanning step, not the same as the distance between the slices)",
    show_default=True,
)  #
@click.option(
    "--angle",
    "-a",
    type=float,
    default=None,
    help="Incident angle of the light sheet, angle between the light sheet and the optical axis in degrees",
    show_default=True,
)  #
@click.option(
    "--flips",
    "-fl",
    type=str,
    default=None,
    help="Flips image to deskew in the opposite orientation (True for view 0 and False for view 1)",
    show_default=True,
)  #
@click.option(
    "--camorientation",
    "-co",
    type=int,
    default=0,
    help="Camera orientation correction expressed as a number of 90 deg rotations to be performed per 2D image in stack -- if required.",
    show_default=True,
)  #
@click.option("--depthaxis", "-za", type=int, default=0, help="Depth axis.", show_default=True)  #
@click.option("--lateralaxis", "-xa", type=int, default=1, help="Lateral axis.", show_default=True)  #
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-1,
    help="Number of worker threads to spawn, if -1 then num workers = num devices",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option(
    "--devices", "-d", type=str, default="0", help="Sets the CUDA devices id, e.g. 0,1,2 or ‘all’", show_default=True
)  #
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def deskew(
    input_paths,
    output_path,
    channels,
    slicing,
    store,
    codec,
    clevel,
    overwrite,
    mode,
    deltax,
    deltaz,
    angle,
    flips,
    camorientation,
    depthaxis,
    lateralaxis,
    workers,
    workersbackend,
    devices,
    check,
):
    """Deskews all or selected channels of a dataset."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_deskew")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    if flips is None:
        flips = (False,) * len(channels)
    else:
        flips = tuple(bool(s) for s in flips.split(","))

    with asection(
        f"Deskewing dataset: {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "
    ):
        aprint(f"Devices used: {devices}, workers: {workers} ")
        dataset_deskew(
            input_dataset,
            output_path,
            channels=channels,
            slicing=slicing,
            dx=deltax,
            dz=deltaz,
            angle=angle,
            flips=flips,
            camera_orientation=camorientation,
            depth_axis=depthaxis,
            lateral_axis=lateralaxis,
            mode=mode,
            store=store,
            compression=codec,
            compression_level=clevel,
            overwrite=overwrite,
            workers=workers,
            workersbackend=workersbackend,
            devices=devices,
            check=check,
        )

        input_dataset.close()
        aprint("Done!")



from typing import Sequence, Tuple

from arbol import aprint, asection
from zarr.errors import ArrayNotFoundError


def dataset_view_remote(
    input_path: str,
    name: str,
    aspect: float,
    channels: Sequence[str],
    contrast_limits: Tuple[float, float],
    colormap: str,
    slicing,
    windowsize: int,
):
    # Annoying napari induced warnings:
    import warnings

    warnings.filterwarnings("ignore")

    if channels is None:
        aprint("Channel(s) must be specified!")
        return

    if ":" not in input_path.replace("http://", ""):
        input_path = f"{input_path}:8000"

    with asection(f"Viewing remote dataset at: {input_path}, channel(s): {channels}"):
        channels = tuple(channel.strip() for channel in channels.split(","))
        channels = list(set(channels))

        from napari import Viewer, gui_qt

        with gui_qt():
            viewer = Viewer(title=f"DEXP | viewing with napari: {name} ", ndisplay=2)
            viewer.window.resize(windowsize + 256, windowsize)

            for channel in channels:
                import dask.array as da

                if "/" in channel:
                    array = da.from_zarr(f"{input_path}/{channel}")
                else:
                    array = da.from_zarr(f"{input_path}/{channel}/{channel}")

                if slicing is not None:
                    array = array[slicing]

                layer = viewer.add_image(
                    array,
                    name=channel,
                    visible=True,
                    contrast_limits=contrast_limits,
                    blending="additive",
                    colormap=colormap,
                    attenuation=0.04,
                    rendering="attenuated_mip",
                )

                if aspect is not None:
                    if array.ndim == 3:
                        layer.scale = (aspect, 1, 1)
                    elif array.ndim == 4:
                        layer.scale = (1, aspect, 1, 1)
                    aprint(f"Setting aspect ratio to {aspect} (layer.scale={layer.scale})")

                try:
                    for axis in range(array.ndim):
                        if "/" in channel:
                            proj_array = da.from_zarr(f"{input_path}/{channel}_max{axis}")
                        else:
                            proj_array = da.from_zarr(f"{input_path}/{channel}/{channel}_max{axis}")

                        proj_layer = viewer.add_image(
                            proj_array,
                            name=f"{channel}_max{axis}",
                            visible=True,
                            contrast_limits=contrast_limits,
                            blending="additive",
                            colormap=colormap,
                        )

                        if aspect is not None:
                            if axis == 0:
                                proj_layer.scale = (1, 1)
                            elif axis == 1:
                                proj_layer.scale = (aspect, 1)
                            elif axis == 2:
                                proj_layer.scale = (aspect, 1)

                            aprint(f"Setting aspect ratio for projection (layer.scale={proj_layer.scale})")

                except (KeyError, ArrayNotFoundError):
                    aprint("Warning: could not find projections in dataset!")



import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _parse_channels, _parse_devices, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.register import dataset_register


@click.command()
@click.argument("input_paths", nargs=-1, required=True)
@click.option("--out-model-path", "-o", default="registration_models.txt", show_default=True)
@click.option(
    "--channels",
    "-c",
    default=None,
    help="list of channels for the view in standard order for the microscope type (C0L0, C0L1, C1L0, C1L1,...)",
)
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option(
    "--microscope",
    "-m",
    type=str,
    default="simview",
    help="Microscope objective to use for computing psf, can be: simview or mvsols",
    show_default=True,
)
@click.option(
    "--equalise/--no-equalise",
    "-eq/-neq",
    default=True,
    help="Equalise intensity of views before fusion, or not.",
    show_default=True,
)
@click.option(
    "--zero-level",
    "-zl",
    type=int,
    default=0,
    help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)",
    show_default=True,
)
@click.option(
    "--clip-high",
    "-ch",
    type=int,
    default=0,
    help="Clips voxel values above the given value, if zero no clipping is done",
    show_default=True,
)
@click.option("--fusion", "-f", type=str, default="tg", help="Fusion mode, can be: ‘tg’ or ‘dct’.  ", show_default=True)
@click.option(
    "--fusion_bias_strength",
    "-fbs",
    type=float,
    default=0.5,
    help="Fusion bias strength for illumination",
    show_default=True,
)
@click.option(
    "--dehaze_size",
    "-dhs",
    type=int,
    default=65,
    help="Filter size (scale) for dehazing the final regsitered and fused image to reduce effect of scattered and out-of-focus light. Set to zero to deactivate.",
    show_default=True,
)
@click.option(
    "--edge-filter",
    "-ef",
    is_flag=True,
    help="Use this flag to apply an edge filter to help registration.",
    show_default=True,
)
@click.option(
    "--max-proj/--no-max-proj",
    "-mp/-nmp",
    type=bool,
    default=True,
    help="Registers using only the maximum intensity projection from each stack.",
    show_default=True,
)
@click.option(
    "--devices", "-d", type=str, default="0", help="Sets the CUDA devices id, e.g. 0,1,2 or ‘all’", show_default=True
)
@click.option(
    "--white-top-hat-size",
    "-wth",
    default=0,
    type=float,
    help="Area opening value after down sampling for white top hat transform transform. Larger values will keep larger components. Recommended value of 1e5.",
)
@click.option(
    "--white-top-hat-sampling", "-wths", default=4, type=int, help="Down sampling size to compute the area opening"
)
@click.option(
    "--remove-beads",
    "-rb",
    is_flag=True,
    default=False,
    help="Use this flag to remove beads before equalizing and fusing",
)
def register(
    input_paths,
    out_model_path,
    channels,
    slicing,
    microscope,
    equalise,
    zero_level,
    clip_high,
    fusion,
    fusion_bias_strength,
    dehaze_size,
    edge_filter,
    max_proj,
    white_top_hat_size,
    white_top_hat_sampling,
    remove_beads,
    devices,
):
    """
    Computes registration model for fusing.
    """
    input_dataset, input_paths = glob_datasets(input_paths)
    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    with asection(
        f"Fusing dataset: {input_paths}, saving it at: {out_model_path}, for channels: {channels}, slicing: {slicing} "
    ):
        aprint(f"Microscope type: {microscope}, fusion type: {fusion}")
        aprint(f"Devices used: {devices}")
        dataset_register(
            dataset=input_dataset,
            model_path=out_model_path,
            channels=channels,
            slicing=slicing,
            microscope=microscope,
            equalise=equalise,
            zero_level=zero_level,
            clip_too_high=clip_high,
            fusion=fusion,
            fusion_bias_strength_i=fusion_bias_strength,
            dehaze_size=dehaze_size,
            registration_edge_filter=edge_filter,
            white_top_hat_size=white_top_hat_size,
            white_top_hat_sampling=white_top_hat_sampling,
            remove_beads=remove_beads,
            max_proj=max_proj,
            devices=devices,
        )



from typing import Optional, Sequence, Tuple

import numpy
from arbol import aprint
from scipy.optimize import minimize

from dexp.utils import xpArray
from dexp.utils.backends import Backend


def linsolve(
    a: xpArray,
    y: xpArray,
    x0: Optional[xpArray] = None,
    maxiter: int = 1e12,
    maxfun: int = 1e12,
    tolerance: float = 1e-6,
    order_error: float = 1,
    order_reg: float = 1,
    alpha_reg: float = 1e-1,
    l2_init: bool = False,
    bounds: Optional[Sequence[Tuple[float, float]]] = None,
    limited: bool = True,
    verbose: bool = False,
) -> xpArray:
    xp = Backend.get_xp_module()

    a = Backend.to_backend(a)
    y = Backend.to_backend(y)

    if x0 is None:
        if l2_init:
            x0 = linsolve(a, y, x0=x0, maxiter=maxiter, tolerance=tolerance, order_error=2, alpha_reg=0, l2_init=False)
        else:
            x0 = numpy.zeros(a.shape[1])

    beta = (1.0 / y.shape[0]) ** (1.0 / order_error)
    alpha = (1.0 / x0.shape[0]) ** (1.0 / order_reg)

    def fun(x):
        x = Backend.to_backend(x)
        if alpha_reg == 0:
            objective = beta * float(xp.linalg.norm(a @ x - y, ord=order_error))
            # aprint(f"objective={objective}, regterm=N/A ")
        else:
            objective = beta * float(xp.linalg.norm(a @ x - y, ord=order_error))
            regularisation_term = (alpha_reg * alpha) * float(xp.linalg.norm(x, ord=order_reg))
            # aprint(f"objective={objective}, regterm={regularisation_term} ")
            objective += regularisation_term
        return objective

    result = minimize(
        fun,
        x0,
        method="L-BFGS-B" if limited else "BFGS",
        tol=tolerance,
        bounds=bounds if limited else None,
        options={
            "disp": verbose,
            "maxiter": maxiter,
            "maxfun": maxfun,
            "gtol": tolerance,
            "eps": 1e-5,  # the minimization would not converge sometimes without this param.
        },
    )

    if result.nit == 0:
        aprint(f"Warning: optimisation finished after {result.nit} iterations!")

    if not result.success:
        raise RuntimeError(
            f"Convergence failed: '{result.message}' after {result.nit} "
            + "iterations and {result.nfev} function evaluations."
        )

    return Backend.to_backend(result.x)



import click

from dexp.cli.defaults import DEFAULT_WORKERS_BACKEND
from dexp.video.crop_resize_pad import crop_resize_pad_image_sequence


@click.command()
@click.argument("input_path", type=str)
@click.option("--output_path", "-o", type=str, default=None, help="Output folder for resized frames.")
@click.option(
    "--crop",
    "-c",
    type=str,
    default=None,
    help="Crop image by removing a given number of pixels/voxels per axis."
    " For example: 10,20*10,20 crops 10 pixels on the left for axis 0, 20 pixels"
    " from the right of axis #0, and the same for axis 2.",
    show_default=True,
)
@click.option(
    "--resize",
    "-r",
    type=str,
    default=None,
    help="After cropping, the image is resized to the given shape."
    " If any entry in the tuple is -1 then that position in the shape"
    " is automatically determined based on the existing shape to preserve aspect ratio.",
    show_default=True,
)
@click.option(
    "--resizeorder",
    "-ro",
    type=int,
    default=3,
    help="The order of the spline interpolation, default is 3. The order has to be in the range 0-5.",
    show_default=True,
)
@click.option(
    "--resizemode",
    "-rm",
    type=str,
    default="constant",
    help="Optional The mode parameter determines how the input array"
    " is extended beyond its boundaries. Can be: ‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’.",
    show_default=True,
)
@click.option(
    "--padwidth",
    "-pw",
    type=str,
    default=None,
    help="After cropping and resizing, padding is performed."
    " The provided tuple is interpreted similarly to cropping."
    " For example, 5,5*5,5 pads all sides with 5 pixels.",
    show_default=True,
)
@click.option(
    "--padmode",
    "-pm",
    type=str,
    default="constant",
    help="Padding mode. See numpy.pad for the available modes",
    show_default=True,
)
@click.option(
    "--padcolor",
    "-pc",
    type=str,
    default="0,0,0,0",
    help="Padding color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.",
    show_default=True,
)
@click.option("--overwrite", "-w", is_flag=True, help="Force overwrite of output images.", show_default=True)
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-1,
    help="Number of worker threads to spawn, set to -1 for maximum number of workers",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option("--device", "-d", type=int, default=0, help="Sets the CUDA devices id, e.g. 0,1,2", show_default=True)  #
def resize(
    input_path,
    output_path,
    crop,
    resize,
    resizeorder,
    resizemode,
    padwidth,
    padmode,
    padcolor,
    overwrite,
    workers,
    workersbackend,
    device,
):
    """Resizes video"""

    # Default output path:
    if output_path is None:
        output_path = input_path + "_resized"
    elif output_path.startswith("_"):
        output_path = input_path + output_path

    if crop is not None:
        crop = tuple(tuple(int(v) for v in xy.split(",")) for xy in crop.split("*"))

    if resize is not None and "," in resize:
        resize = tuple(int(v) for v in resize.split(","))

    if padwidth is not None:
        padwidth = tuple(tuple(int(v) for v in xy.split(",")) for xy in padwidth.split("*"))

    # Parse color:
    if padcolor is not None and "," in padcolor:
        padcolor = tuple(float(v) for v in padcolor.split(","))

    crop_resize_pad_image_sequence(
        input_path=input_path,
        output_path=output_path,
        crop=crop,
        resize=resize,
        resize_order=resizeorder,
        resize_mode=resizemode,
        pad_width=padwidth,
        pad_mode=padmode,
        pad_color=padcolor,
        overwrite=overwrite,
        workers=workers,
        workersbackend=workersbackend,
        device=device,
    )



import itertools
import math
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
from arbol import aprint, asection
from scipy.optimize import minimize, shgo

from dexp.processing.denoising.metrics import mean_squared_error
from dexp.utils import dict_or, xpArray
from dexp.utils.backends import Backend


def calibrate_denoiser(
    image: xpArray,
    denoise_function: Callable[[Any], xpArray],
    denoise_parameters: Dict[str, List[Union[float, int]]],
    setup_function: Optional[Callable[[xpArray], Any]] = None,
    mode: str = "shgo+lbfgs",
    max_evaluations: int = 4096,
    stride: int = 4,
    loss_function: Callable = mean_squared_error,
    display: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates denoiser using self-supervised loss from Batson & Royer*
    Derived from code here:
    https://scikit-image.org/docs/dev/auto_examples/filters/plot_j_invariant_tutorial.html
    Reference: "Noise2Self: Blind Denoising by Self-Supervision, International
    Conference on Machine Learning, p. 524-533 (2019)"

    This 'classic_denoisers' version uses a 'brute-force' optimizer. Good when the
    denoiser is fast enough and the parameter space to explore small enough.

    Parameters
    ----------
    image: ArrayLike
        Image to calibate denoiser with.
    denoise_function: Callable
        Denosing function to calibrate. Should take an image as first parameter,
        all other parameters should have defaults
    denoise_parameters:
        Dictionary with keys corresponding to parameters of the denoising function.
        Values are either: (i) a list of possible values (categorical parameter),
        or (ii) a tuple of floats defining the bounds of that numerical parameter.
    setup_function : Callable, optional
        Function to pre process / setup denoising input.
    mode : str
        Optimisation mode. Can be: 'bruteforce', 'lbfgs' or 'shgo'.
    max_evaluations: int
        Maximum number of function evaluations during optimisation.
    stride: int
        Stride to compute self-supervised loss.
    loss_function: Callable
        Loss/Error function: takes two arrays and returns a distance-like function.
        Can be:  structural_error, mean_squared_error, _mean_absolute_error
    display_images: bool
        If True the denoised images for each parameter tested are displayed.
        this _will_ be slow.
    other_fixed_parameters: dict
        Other fixed parameters to pass to the denoiser function.


    Returns
    -------
    Dictionary with optimal parameters

    """
    # Move image to backend:
    image = Backend.to_backend(image)

    aprint(f"Calibrating denoiser on image of shape: {image.shape}")
    aprint(f"Stride for Noise2Self loss: {stride}")
    aprint(f"Fixed parameters: {other_fixed_parameters}")

    # Pass fixed parameters:
    denoise_function = partial(denoise_function, **other_fixed_parameters)

    with asection(f"Calibrating denoiser with method: {mode}"):
        best_parameters = _calibrate_denoiser_search(
            image,
            denoise_function,
            denoise_parameters=denoise_parameters,
            setup_function=setup_function,
            mode=mode,
            max_evaluations=max_evaluations,
            stride=stride,
            loss_function=loss_function,
            display_images=display,
        )

    aprint(f"Best parameters are: {best_parameters}")

    return dict_or(best_parameters, other_fixed_parameters)


def _interpolate_image(image: xpArray):

    # Backend:
    sp = Backend.get_sp_module(image)

    conv_filter = sp.ndimage.generate_binary_structure(image.ndim, 1).astype(image.dtype)
    conv_filter.ravel()[conv_filter.size // 2] = 0
    conv_filter /= conv_filter.sum()

    interpolation = sp.ndimage.convolve(image, conv_filter, mode="mirror")

    return interpolation


def _generate_mask(image: xpArray, stride: int = 4):

    # Generate slice for mask:
    spatialdims = image.ndim
    n_masks = stride ** spatialdims
    phases = np.unravel_index(n_masks // 2, (stride,) * len(image.shape[:spatialdims]))
    mask = tuple(slice(p, None, stride) for p in phases)

    return mask


def _product_from_dict(dictionary: Dict[str, List[Union[float, int]]]):
    """Utility function to convert parameter ranges to parameter combinations.

    Converts a dict of lists into a list of dicts whose values consist of the
    cartesian product of the values in the original dict.

    Parameters
    ----------
    dictionary : dict of lists
        Dictionary of lists to be multiplied.

    Yields
    ------
    selections : dicts of values
        Dicts containing individual combinations of the values in the input
        dict.
    """
    keys = dictionary.keys()
    for element in itertools.product(*dictionary.values()):
        yield dict(zip(keys, element))


def _calibrate_denoiser_search(
    image: xpArray,
    denoise_function: Callable[[Any], xpArray],
    denoise_parameters: Dict[str, List[Union[float, int]]],
    setup_function: Optional[Callable[[xpArray], Any]],
    mode: str,
    max_evaluations: int,
    stride=4,
    loss_function: Callable = mean_squared_error,  # _structural_loss, #
    display_images: bool = False,
):
    """Return a parameter search history with losses for a denoise function.

    Parameters
    ----------
    image : ndarray
        Input data to be denoised (converted using `img_as_float`).
    denoise_function : Callable
        Denoising function to be calibrated.
    denoise_parameters : dict of list
        Ranges of parameters for `denoise_function` to be calibrated over.
    setup_function : Callable, optional
        Function to pre process / setup denoising input.
    mode : str
        Optimisation mode. Can be: "bruteforce", "lbfgs" or "shgo".
    max_evaluations: int
        Maximum number of function evaluations during optimisation.
    stride : int, optional
        Stride used in masking procedure that converts `denoise_function`
        to J-invariance.
    loss_function : Callable
        Loss function to use
    display : bool
        When True the resulting images are displayed with napari

    Returns
    -------
    parameters_tested : list of dict
        List of parameters tested for `denoise_function`, as a dictionary of
        kwargs.
    losses : list of int
        Self-supervised loss for each set of parameters in `parameters_tested`.
    """

    # Move image to backend:
    image = Backend.to_backend(image)

    # Generate mask:
    mask = _generate_mask(image, stride)
    masked_image = image.copy()
    masked_image[mask] = _interpolate_image(image)[mask]

    # denoised images are kept here:
    denoised_images = []

    # Parameter names:
    parameter_names = list(denoise_parameters.keys())

    # Best parameters (to be found):
    best_parameters = None

    # Setting up denoising
    if setup_function is not None:
        denoising_input = setup_function(masked_image)
    else:
        denoising_input = masked_image

    # Function to optimise:
    def _loss_func(**_denoiser_kwargs):
        # We compute the J-inv loss:
        denoised = denoise_function(denoising_input, **_denoiser_kwargs)
        loss = loss_function(denoised[mask], image[mask])

        if math.isnan(loss) or math.isinf(loss):
            loss = math.inf

        aprint(f"J-inv loss is: {loss}")

        loss = Backend.to_numpy(loss)

        if display_images and not (math.isnan(loss) or math.isinf(loss)):
            denoised = denoise_function(image, **_denoiser_kwargs)
            denoised_images.append(denoised)

        return -float(loss)

    best_parameters = None

    if "bruteforce" in mode:
        with asection(f"Searching by brute-force for the best denoising parameters among: {denoise_parameters}"):

            num_rounds = 4
            best_loss = -math.inf
            for _ in range(num_rounds):
                # expand ranges:
                expanded_denoise_parameters = {n: np.arange(*r) for (n, r) in denoise_parameters.items()}
                # Generate all possible combinations:
                cartesian_product_of_parameters = list(_product_from_dict(expanded_denoise_parameters))
                for denoiser_kwargs in cartesian_product_of_parameters:
                    with asection(f"computing J-inv loss for: {denoiser_kwargs}"):
                        loss = _loss_func(**denoiser_kwargs)
                        if loss > best_loss:
                            best_loss = loss
                            best_parameters = denoiser_kwargs

    if "shgo" in mode:
        with asection(
            "Searching by 'simplicial homology global optimization' (SHGO)"
            f"the best denoising parameters among: {denoise_parameters}"
        ):
            if best_parameters is None:
                x0 = tuple(0.5 * (v[1] - v[0]) for v in denoise_parameters.values())
            else:
                x0 = tuple(best_parameters[k] for k in denoise_parameters.keys())

            bounds = list(v[0:2] for v in denoise_parameters.values())

            # Impedance mismatch:
            def _func(*_denoiser_kwargs):
                param_dict = {n: v for (n, v) in zip(parameter_names, tuple(_denoiser_kwargs[0]))}
                value = -_loss_func(**param_dict)
                return value

            result = shgo(_func, bounds, sampling_method="sobol", options={"maxev": max_evaluations})
            aprint(result)
            best_parameters = dict({n: v for (n, v) in zip(parameter_names, result.x)})

    if "lbfgs" in mode:
        with asection(f"Searching by 'Limited-memory BFGS' the best denoising parameters among: {denoise_parameters}"):

            if best_parameters is None:
                x0 = tuple(0.5 * (v[1] - v[0]) for v in denoise_parameters.values())
            else:
                x0 = tuple(best_parameters[k] for k in denoise_parameters.keys())

            bounds = list(v[0:2] for v in denoise_parameters.values())

            # Impedance mismatch:
            def _func(*_denoiser_kwargs):
                param_dict = {n: v for (n, v) in zip(parameter_names, tuple(_denoiser_kwargs[0]))}
                value = -_loss_func(**param_dict)
                return value

            result = minimize(
                fun=_func,
                x0=x0,
                method="L-BFGS-B",
                bounds=bounds,
                options=dict(maxfun=max_evaluations, eps=1e-2, ftol=1e-9, gtol=1e-9),
            )
            aprint(result)
            best_parameters = dict({n: v for (n, v) in zip(parameter_names, result.x)})

    if display_images:
        import napari

        viewer = napari.Viewer()
        viewer.add_image(Backend.to_numpy(image), name="image")
        viewer.add_image(np.stack([Backend.to_numpy(i) for i in denoised_images]), name="denoised")
        napari.run()

    return best_parameters



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import (
    DEFAULT_CLEVEL,
    DEFAULT_CODEC,
    DEFAULT_STORE,
    DEFAULT_WORKERS_BACKEND,
)
from dexp.cli.parsing import (
    _get_output_path,
    _parse_channels,
    _parse_devices,
    _parse_slicing,
)
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.deconv import dataset_deconv


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="list of channels, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True)
@click.option("--tilesize", "-ts", type=int, default=512, help="Tile size for tiled computation", show_default=True)
@click.option(
    "--method",
    "-m",
    type=str,
    default="lr",
    help="Deconvolution method: for now only lr (Lucy Richardson)",
    show_default=True,
)
@click.option(
    "--iterations",
    "-i",
    type=int,
    default=None,
    help="Number of deconvolution iterations. More iterations takes longer, will be sharper, but might also be potentially more noisy depending on method. "
    "The default number of iterations depends on the other parameters, in particular it depends on the choice of backprojection operator. For ‘wb’ as little as 3 iterations suffice. ",
    show_default=True,
)
@click.option(
    "--maxcorrection",
    "-mc",
    type=int,
    default=None,
    help="Max correction in folds per iteration. By default there is no limit",
    show_default=True,
)
@click.option(
    "--power",
    "-pw",
    type=float,
    default=1.0,
    help="Correction exponent, default for standard LR is 1, set to >1 for acceleration.",
    show_default=True,
)
@click.option(
    "--blindspot",
    "-bs",
    type=int,
    default=0,
    help="Blindspot based noise reduction. Provide size of kernel to use, must be an odd number: 3(recommended), 5, 7. 0 means no blindspot. ",
    show_default=True,
)
@click.option(
    "--backprojection",
    "-bp",
    type=str,
    default="tpsf",
    help="Back projection operator, can be: ‘tpsf’ (transposed PSF = classic) or ‘wb’ (Wiener-Butterworth =  accelerated) ",
    show_default=True,
)
@click.option(
    "--wb-order",
    "-wbo",
    type=int,
    default=5,
    help="Wiener-Butterworth order parameter, a higher `n` makes the filter transition slope closer to a hard cutoff, causing the ringing artifacts in the spatial domain."
    "In contrast, a lower order `n` makes the transition slope gentler and sacrifice some spectral amplitude at spatial frequencies close to the cutoff.",
    show_default=True,
)
@click.option(
    "--objective",
    "-obj",
    type=str,
    default="nikon16x08na",
    help="Microscope objective to use for computing psf, can be: nikon16x08na, olympus20x10na, or path to file",
    show_default=True,
)
@click.option("--numape", "-na", type=float, default=None, help="Overrides NA value for objective.", show_default=True)
@click.option("--dxy", "-dxy", type=float, default=0.485, help="Voxel size along x and y in microns", show_default=True)
@click.option("--dz", "-dz", type=float, default=4 * 0.485, help="Voxel size along z in microns", show_default=True)
@click.option("--xysize", "-sxy", type=int, default=31, help="PSF size along xy in voxels", show_default=True)
@click.option("--zsize", "-sz", type=int, default=31, help="PSF size along z in voxels", show_default=True)
@click.option("--showpsf", "-sp", is_flag=True, help="Show point spread function (PSF) with napari", show_default=True)
@click.option(
    "--scaling",
    "-sc",
    type=str,
    default="1,1,1",
    help="Scales input image along the three axis: sz,sy,sx (numpy order). For example: 2,1,1 upscales along z by a factor 2",
    show_default=True,
)  #
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-1,
    help="Number of worker threads to spawn, if -1 then num workers = num devices",
    show_default=True,
)
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option(
    "--devices", "-d", type=str, default="0", help="Sets the CUDA devices id, e.g. 0,1,2 or ‘all’", show_default=True
)  #
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def deconv(
    input_paths,
    output_path,
    channels,
    slicing,
    store,
    codec,
    clevel,
    overwrite,
    tilesize,
    method,
    iterations,
    maxcorrection,
    power,
    blindspot,
    backprojection,
    wb_order,
    objective,
    numape,
    dxy,
    dz,
    xysize,
    zsize,
    showpsf,
    scaling,
    workers,
    workersbackend,
    devices,
    check,
):
    """Deconvolves all or selected channels of a dataset."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_deconv")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    if "," in scaling:
        scaling = tuple(float(v) for v in scaling.split(","))

    with asection(
        f"Deconvolving dataset: {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "
    ):
        dataset_deconv(
            input_dataset,
            output_path,
            channels=channels,
            slicing=slicing,
            store=store,
            compression=codec,
            compression_level=clevel,
            overwrite=overwrite,
            tilesize=tilesize,
            method=method,
            num_iterations=iterations,
            max_correction=maxcorrection,
            power=power,
            blind_spot=blindspot,
            back_projection=backprojection,
            wb_order=wb_order,
            psf_objective=objective,
            psf_na=numape,
            psf_dxy=dxy,
            psf_dz=dz,
            psf_xy_size=xysize,
            psf_z_size=zsize,
            psf_show=showpsf,
            scaling=scaling,
            workers=workers,
            workersbackend=workersbackend,
            devices=devices,
            check=check,
        )

        input_dataset.close()
        aprint("Done!")



# flake8: noqa

from arbol import Arbol
from skimage import data
from skimage.color import rgb2gray

from dexp.processing.denoising.gaussian import calibrate_denoise_gaussian
from dexp.processing.denoising.metrics import psnr, ssim
from dexp.processing.denoising.noise import add_noise
from dexp.utils.backends import Backend, CupyBackend, NumpyBackend


def demo_gaussian_numpy():
    with NumpyBackend():
        _demo_gaussian()


def demo_gaussian_cupy():
    try:
        with CupyBackend():
            _demo_gaussian()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_gaussian(display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    # Backend:
    xp = Backend.get_xp_module()

    Arbol.enable_output = True
    Arbol.set_log_max_depth(5)

    image = data.astronaut()
    image = rgb2gray(image)
    image = Backend.to_backend(image)

    noisy = add_noise(image)

    function, parameters = calibrate_denoise_gaussian(noisy)
    denoised = function(noisy, **parameters)

    image = xp.clip(image, 0, 1)
    noisy = xp.clip(noisy, 0, 1)
    denoised = xp.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("         noisy   :", psnr_noisy, ssim_noisy)
    print("gaussian denoised:", psnr_denoised, ssim_denoised)

    if display:
        import napari

        viewer = napari.Viewer()
        viewer.add_image(Backend.to_numpy(image), name="image")
        viewer.add_image(Backend.to_numpy(noisy), name="noisy")
        viewer.add_image(Backend.to_numpy(denoised), name="denoised")
        napari.run()

    return ssim_denoised


if __name__ == "__main__":
    demo_gaussian_cupy()
    demo_gaussian_numpy()



from typing import Any, Generator, Sequence, Tuple, Union

import numpy

from dexp.utils.backends import Backend


def blend_color_images(
    images: Union[Generator[Any, Any, None], Sequence[Any]],
    modes: Union[str, Sequence[str]] = "max",
    alphas: Sequence[float] = None,
    background_color: Tuple[float, float, float, float] = (0, 0, 0, 0),
    rgba_value_max: float = 255,
    internal_dtype=numpy.float32,
):
    """
    Blends multiple images together according to a blend mode

    Parameters
    ----------
    images: Sequence of images to blend.
    modes: Blending modes. Either one for all images, or one per image in the form of a sequence.
        Blending modes are: 'mean', 'add', 'satadd', 'max', 'alpha'
    alphas: Optional sequence of alpha values to use for blending, must be of same length as
        the sequence of images, optional.
    background_color: Background color as tuple of normalised floats:  (R,G,B,A). Default is transparent black.
    rgba_value_max: Max value for the pixel/voxel values.
    internal_dtype: dtype for internal computation

    Returns
    -------
    Blended image

    """

    xp = Backend.get_xp_module()

    # convert to list of images, and move to backend:
    images = list(Backend.to_backend(image) for image in images)

    # Check that there is at least one image in the image list:
    if len(images) == 0:
        raise ValueError("Blending requires at least one image!")

    # original dtype:
    original_dtype = images[0].dtype

    # Verify that all images are of the same shape:
    for image in images:
        if images[0].ndim != image.ndim or images[0].shape[:-1] != image.shape[:-1]:
            raise ValueError("All images in sequence must have the same number of dimensions and shape!")

    # verify that the images are 2D RGB(A) images:
    if images[0].ndim != 3 or not images[0].shape[2] in (3, 4):
        raise ValueError("Images must be 2D RGB(A) images!")

    # Check that the number of images and alphas match:
    if len(images) != len(alphas):
        raise ValueError(
            f"The number of images ({len(images)}) is not equal to the number of alpha values ({len(images)})."
        )

    # Create image to store result:
    result = xp.zeros_like(_ensure_rgba(images[0], rgba_value_max), dtype=internal_dtype)

    # Fill with background color:
    result[..., :] = xp.asarray(background_color, dtype=internal_dtype)

    # Blend in the images:
    for image, mode, alpha in zip(images, modes, alphas):
        # move image to backend:
        image = Backend.to_backend(image, dtype=internal_dtype)

        image = _ensure_rgba(image, rgba_value_max)

        # normalise image:
        image /= rgba_value_max
        result = _blend_function(alpha * image, result, mode)

    # scale back:
    result *= rgba_value_max

    # clip:
    result = xp.clip(result, 0, rgba_value_max)

    # convert back to original dtype (of first image)
    result = result.astype(dtype=original_dtype, copy=False)

    return result


def _ensure_rgba(image, value_max):
    xp = Backend.get_xp_module()
    # make sure that the image has an alpha channel:
    if image.shape[-1] == 3:
        pad_width = ((0, 0),) * (image.ndim - 1) + ((0, 1),)
        image = xp.pad(image, pad_width=pad_width)
        image[..., -1] = value_max
    return image


def _blend_function(image_u, image_v, mode):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if mode == "mean":
        result = 0.5 * (image_u + image_v)

    elif mode == "satadd":
        # TODO: this fails but it is unclear why:
        result = sp.special.erf(image_u + image_v)

    elif mode == "max" or mode == "rgbamax":
        result = xp.maximum(image_u, image_v)

    elif mode == "min" or mode == "rgbmin":
        result = xp.minimum(image_u, image_v)
        result[..., 3] = xp.maximum(image_u[..., 3], image_v[..., 3])

    elif image_u.shape[-1] == 4 and image_v.shape[-1] == 4:
        # See: https://en.wikipedia.org/wiki/Alpha_compositing
        # See: https://www.cairographics.org/operators/
        result = xp.zeros_like(image_u)

        src_rgb = image_u[..., 0:3]
        src_alpha = image_u[..., 3][..., xp.newaxis]
        dst_rgb = image_v[..., 0:3]
        dst_alpha = image_v[..., 3][..., xp.newaxis]

        if mode == "clear":
            # Where the second object is drawn, the first is completely removed. Anywhere else it is left intact.
            # The second object itself is not drawn.
            out_alpha = 0
            out_rgb = 0
        elif mode == "source":
            # The second object is drawn as if nothing else were below. Only outside of the blue rectangle
            # the red one is left intact.
            out_alpha = src_alpha
            out_rgb = src_rgb
        elif mode == "over" or mode == "alpha":
            # The image shows what you would expect if you held two semi-transparent slides on top of each other.
            # This operator is cairo's default operator.
            out_alpha = src_alpha + dst_alpha * (1 - src_alpha)
            out_rgb = (src_rgb * src_alpha + dst_rgb * dst_alpha * (1 - src_alpha)) / (out_alpha + 1e-8)
        elif mode == "in":
            # The first object is removed completely, the second is only drawn where the first was.
            out_alpha = src_alpha * dst_alpha
            out_rgb = src_rgb
        elif mode == "out":
            # The blue rectangle is drawn only where the red one wasn't. Since the red one was partially
            # transparent, you can see a blue shadow in the overlapping area. Otherwise,
            # the red object is completely removed.
            out_alpha = src_alpha * (1 - dst_alpha)
            out_rgb = src_rgb
        elif mode == "atop":
            # This leaves the first object mostly intact, but mixes both objects in the overlapping area.
            # The second object object is not drawn except there.
            # If you look closely, you will notice that the resulting color in the overlapping area is different from
            # what the OVER operator produces. Any two operators produce different output in the overlapping area!
            out_alpha = dst_alpha
            out_rgb = src_alpha * src_rgb + dst_rgb * (1 - src_alpha)
        elif mode == "dest":
            # Leaves the first object untouched, the second is discarded completely.
            out_alpha = dst_alpha
            out_rgb = dst_rgb
        elif mode == "dest_over":
            # The result is similar to the OVER operator. Except that the "order" of the objects is reversed,
            # so the second is drawn below the first.
            out_alpha = (1 - dst_alpha) * src_alpha + dst_alpha
            out_rgb = (src_alpha * src_rgb * (1 - dst_alpha) + dst_alpha * dst_rgb) / (out_alpha + 1e-8)
        elif mode == "dest_in":
            # The blue rectangle is used to determine which part of the red one is left intact.
            # Anything outside the overlapping area is removed.
            # This works like the IN operator, but again with the second object "below" the first.
            out_alpha = src_alpha * dst_alpha
            out_rgb = dst_rgb
        elif mode == "dest_atop":
            # Same as the ATOP operator, but again as if the order of the drawing operations
            # had been reversed.
            out_alpha = src_alpha
            out_rgb = src_rgb * (1 - dst_alpha) + dst_alpha * dst_rgb
        elif mode == "xor":
            # The output of the XOR operator is the same for both bounded and unbounded source interpretations.
            out_alpha = src_alpha + dst_alpha - 2 * src_alpha * dst_alpha
            out_rgb = src_alpha * src_rgb * (1 - dst_alpha) + dst_alpha * dst_rgb * (1 - src_alpha)
            out_rgb /= out_alpha + 1e-8
        elif mode == "add":
            # The output of the ADD operator is the same for both bounded and unbounded source interpretations.
            out_alpha = xp.minimum(1, src_alpha + dst_alpha)
            out_rgb = (src_alpha * src_rgb + dst_alpha * dst_rgb) / (out_alpha + 1e-8)
        elif mode == "saturate":
            # The output of the ADD operator is the same for both bounded and unbounded source interpretations.
            out_alpha = xp.minimum(1, src_alpha + dst_alpha)
            out_rgb = (xp.minimum(src_alpha, 1 - dst_alpha) * src_rgb + dst_alpha * dst_rgb) / (out_alpha + 1e-8)
        else:
            raise ValueError(
                f"Invalid alpha blending mode {mode} or incompatible images: {image_u.shape} {image_v.shape}"
            )

        result[..., 0:3] = out_rgb.squeeze()
        result[..., 3] = out_alpha.squeeze()
    else:
        raise ValueError(f"Invalid blending mode {mode} or incompatible images: {image_u.shape} {image_v.shape}")

    return result



# flake8: noqa
from arbol import Arbol

from dexp.datasets.synthetic_datasets import generate_nuclei_background_data
from dexp.processing.denoising.butterworth import calibrate_denoise_butterworth
from dexp.processing.denoising.metrics import psnr, ssim
from dexp.processing.denoising.noise import add_noise
from dexp.utils.backends import Backend, CupyBackend, NumpyBackend
from dexp.utils.timeit import timeit


def demo_butterworth_numpy():
    with NumpyBackend():
        _demo_butterworth()


def demo_butterworth_cupy():
    try:
        with CupyBackend():
            _demo_butterworth()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_butterworth(display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    # Backend:
    xp = Backend.get_xp_module()

    Arbol.enable_output = True
    Arbol.set_log_max_depth(5)

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=True, length_xy=320, length_z_factor=1)

    image = Backend.to_backend(image, dtype=xp.float32)
    image /= image.max()

    noisy = add_noise(image)

    function, parameters = calibrate_denoise_butterworth(noisy, mode="xy-z")
    denoised = function(noisy, **parameters)

    image = xp.clip(image, 0, 1)
    noisy = xp.clip(noisy, 0, 1)
    denoised = xp.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("        noisy   :", psnr_noisy, ssim_noisy)
    print("lowpass denoised:", psnr_denoised, ssim_denoised)

    if display:
        import napari

        viewer = napari.Viewer()
        viewer.add_image(Backend.to_numpy(image), name="image")
        viewer.add_image(Backend.to_numpy(noisy), name="noisy")
        viewer.add_image(Backend.to_numpy(denoised), name="denoised")
        napari.run()

    return ssim_denoised


if __name__ == "__main__":
    demo_butterworth_cupy()
    demo_butterworth_numpy()



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import DEFAULT_CLEVEL, DEFAULT_CODEC, DEFAULT_STORE
from dexp.cli.parsing import (
    _get_output_path,
    _parse_channels,
    _parse_devices,
    _parse_slicing,
)
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.fuse import dataset_fuse


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option(
    "--channels",
    "-c",
    default=None,
    help="list of channels for the view in standard order for the microscope type (C0L0, C0L1, C1L0, C1L1,...)",
)
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)  #
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="compression codec: ‘zstd’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option(
    "--microscope",
    "-m",
    type=str,
    default="simview",
    help="Microscope objective to use for computing psf, can be: simview or mvsols",
    show_default=True,
)
@click.option(
    "--equalise/--no-equalise",
    "-eq/-neq",
    default=True,
    help="Equalise intensity of views before fusion, or not.",
    show_default=True,
)
@click.option(
    "--equalisemode",
    "-eqm",
    default="first",
    help="Equalisation modes: compute correction ratios only for first time point: ‘first’ or for all time points: ‘all’.",
    show_default=True,
)
@click.option(
    "--zerolevel",
    "-zl",
    type=int,
    default=0,
    help="‘zero-level’ i.e. the pixel values in the restoration (to be substracted)",
    show_default=True,
)  #
@click.option(
    "--cliphigh",
    "-ch",
    type=int,
    default=0,
    help="Clips voxel values above the given value, if zero no clipping is done",
    show_default=True,
)  #
@click.option(
    "--fusion", "-f", type=str, default="tg", help="Fusion mode, can be: ‘tg’ or ‘dct’.  ", show_default=True
)  #
@click.option(
    "--fusion_bias_strength",
    "-fbs",
    type=(float, float),
    default=(0.5, 0.02),
    help="Fusion bias strength for illumination and detection ‘fbs_i fbs_d’, set to ‘0 0’) if fusing a cropped region",
    show_default=True,
)  #
@click.option(
    "--dehaze_size",
    "-dhs",
    type=int,
    default=65,
    help="Filter size (scale) for dehazing the final regsitered and fused image to reduce effect of scattered and out-of-focus light. Set to zero to deactivate.",
    show_default=True,
)  #
@click.option(
    "--dark-denoise-threshold",
    "-ddt",
    type=int,
    default=0,
    help="Threshold for denoises the dark pixels of the image -- helps increase compression ratio. Set to zero to deactivate.",
    show_default=True,
)  #
@click.option(
    "--zpadapodise",
    "-zpa",
    type=(int, int),
    default=(8, 96),
    help="Pads and apodises the views along z before fusion: ‘pad apo’, where pad is a padding length, and apo is apodisation length, both in voxels. If pad=apo, no original voxel is modified and only added voxels are apodised.",
    show_default=True,
)  #
@click.option(
    "--loadreg",
    "-lr",
    is_flag=True,
    help="Turn on to load the registration parameters from a previous run",
    show_default=True,
)  #
@click.option(
    "--model-filename",
    "-mf",
    help="Model filename to load or save registration model list",
    default="registration_models.txt",
    show_default=True,
)
@click.option(
    "--warpregiter",
    "-wri",
    type=int,
    default=4,
    help="Number of iterations for warp registration (if applicable).",
    show_default=True,
)  #
@click.option(
    "--minconfidence",
    "-mc",
    type=float,
    default=0.3,
    help="Minimal confidence for registration parameters, if below that level the registration parameters for previous time points is used.",
    show_default=True,
)  #
@click.option(
    "--maxchange",
    "-md",
    type=float,
    default=16,
    help="Maximal change in registration parameters, if above that level the registration parameters for previous time points is used.",
    show_default=True,
)  #
@click.option(
    "--regedgefilter",
    "-ref",
    is_flag=True,
    help="Use this flag to apply an edge filter to help registration.",
    show_default=True,
)  #
@click.option(
    "--maxproj/--no-maxproj",
    "-mp/-nmp",
    type=bool,
    default=True,
    help="Registers using only the maximum intensity projection from each stack.",
    show_default=True,
)
@click.option(
    "--hugedataset",
    "-hd",
    is_flag=True,
    help="Use this flag to indicate that the the dataset is _huge_ and that memory allocation should be optimised at the detriment of processing speed.",
    show_default=True,
)  #
@click.option(
    "--devices", "-d", type=str, default="0", help="Sets the CUDA devices id, e.g. 0,1,2 or ‘all’", show_default=True
)  #
@click.option(
    "--pad", "-p", is_flag=True, default=False, help="Use this flag to pad views according to the registration models."
)
@click.option(
    "--white-top-hat-size",
    "-wth",
    default=0,
    type=float,
    help="Area opening value after down sampling for white top hat transform transform. Larger values will keep larger components. Recommended value of 1e5.",
)
@click.option(
    "--white-top-hat-sampling", "-wths", default=4, type=int, help="Down sampling size to compute the area opening"
)
@click.option(
    "--remove-beads",
    "-rb",
    is_flag=True,
    default=False,
    help="Use this flag to remove beads before equalizing and fusing",
)
@click.option("--check", "-ck", default=True, help="Checking integrity of written file.", show_default=True)  #
def fuse(
    input_paths,
    output_path,
    channels,
    slicing,
    store,
    codec,
    clevel,
    overwrite,
    microscope,
    equalise,
    equalisemode,
    zerolevel,
    cliphigh,
    fusion,
    fusion_bias_strength,
    dehaze_size,
    dark_denoise_threshold,
    zpadapodise,
    loadreg,
    model_filename,
    warpregiter,
    minconfidence,
    maxchange,
    regedgefilter,
    maxproj,
    hugedataset,
    devices,
    pad,
    white_top_hat_size,
    white_top_hat_sampling,
    remove_beads,
    check,
):
    """Fuses the views of a multi-view light-sheet microscope dataset (available: simview and mvsols)"""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_fused")

    slicing = _parse_slicing(slicing)
    channels = _parse_channels(input_dataset, channels)
    devices = _parse_devices(devices)

    with asection(
        f"Fusing dataset: {input_paths}, saving it at: {output_path}, for channels: {channels}, slicing: {slicing} "
    ):
        aprint(f"Microscope type: {microscope}, fusion type: {fusion}")
        aprint(f"Devices used: {devices}")
        dataset_fuse(
            input_dataset,
            output_path,
            channels=channels,
            slicing=slicing,
            store=store,
            compression=codec,
            compression_level=clevel,
            overwrite=overwrite,
            microscope=microscope,
            equalise=equalise,
            equalise_mode=equalisemode,
            zero_level=zerolevel,
            clip_too_high=cliphigh,
            fusion=fusion,
            fusion_bias_strength_i=fusion_bias_strength[0],
            fusion_bias_strength_d=fusion_bias_strength[1],
            dehaze_size=dehaze_size,
            dark_denoise_threshold=dark_denoise_threshold,
            z_pad_apodise=zpadapodise,
            loadreg=loadreg,
            model_list_filename=model_filename,
            warpreg_num_iterations=warpregiter,
            min_confidence=minconfidence,
            max_change=maxchange,
            registration_edge_filter=regedgefilter,
            maxproj=maxproj,
            huge_dataset=hugedataset,
            devices=devices,
            pad=pad,
            white_top_hat_size=white_top_hat_size,
            white_top_hat_sampling=white_top_hat_sampling,
            remove_beads=remove_beads,
            check=check,
        )

        input_dataset.close()
        aprint("Done!")



from dexp.utils.backends import Backend, CupyBackend


def clear_fft_plan_cache() -> None:
    """Clears fft plan cache
    Reference: https://github.com/cupy/cupy/issues/5134#issuecomment-833619964
    """
    backend = Backend.current()
    if not isinstance(backend, CupyBackend):
        return

    from cupy.fft.config import get_plan_cache

    cache = get_plan_cache()
    prev_size = cache.get_size()
    prev_mem = cache.get_memsize()
    cache.set_size(0)

    backend.clear_memory_pool()

    cache.set_size(prev_size)
    cache.set_memsize(prev_mem)



from dexp.datasets.synthetic_datasets.binary_blobs import binary_blobs
from dexp.datasets.synthetic_datasets.multiview_data import generate_fusion_test_data
from dexp.datasets.synthetic_datasets.nuclei_background_data import (
    generate_nuclei_background_data,
)



import click

from dexp.cli.parsing import _parse_channels, _parse_slicing
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.histogram import dataset_histogram


@click.command()
@click.argument("input-paths", nargs=-1)
@click.option("--output-directory", "-o", default="histogram", show_default=True, type=str)
@click.option("--channels", "-c", default=None, help="list of channels, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option("--device", "-d", type=int, default=0, show_default=True)
@click.option(
    "--minimum-count",
    "-m",
    default=5,
    show_default=True,
    type=int,
    help="Minimum count on the histogram used to select upper limit, values lower than this are not considered.",
)
@click.option(
    "--maximum-count",
    "-M",
    default=None,
    show_default=True,
    type=float,
    help="Maximum count to clip histogram and improve visualization.",
)
def histogram(input_paths, output_directory, channels, slicing, device, minimum_count, maximum_count):

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)
    slicing = _parse_slicing(slicing)

    if slicing is not None:
        input_dataset.set_slicing(slicing)

    dataset_histogram(
        dataset=input_dataset,
        output_dir=output_directory,
        channels=channels,
        device=device,
        minimum_count=minimum_count,
        maximum_count=maximum_count,
    )



from arbol.arbol import asection

from dexp.utils.backends import Backend


@asection("Generating synthetic binary blobs data")
def binary_blobs(length=512, blob_size_fraction=0.1, n_dim=2, volume_fraction=0.5, rng=42):
    """
    Generate synthetic binary image with several rounded blob-like objects.

    Parameters
    ----------
    length : int, optional
        Linear size of output image.
    blob_size_fraction : float, optional
        Typical linear size of blob, as a fraction of ``length``, should be
        smaller than 1.
    n_dim : int, optional
        Number of dimensions of output image.
    volume_fraction : float, default 0.5
        Fraction of image pixels covered by the blobs (where the output is 1).
        Should be in [0, 1].
    rng : int or Generator
        If an integer is provided it's used to create an numpy (cupy) default random number generator.

    Returns
    -------
    blobs : ndarray of bools
        Output binary image

    Examples
    --------
    >>> data.binary_blobs(length=5, blob_size_fraction=0.2, seed=1)
    array([[ True, False,  True,  True,  True],
           [ True,  True,  True, False,  True],
           [False,  True, False,  True,  True],
           [ True, False, False,  True,  True],
           [ True, False, False, False,  True]])
    >>> blobs = data.binary_blobs(length=256, blob_size_fraction=0.1)
    >>> # Finer structures
    >>> blobs = data.binary_blobs(length=256, blob_size_fraction=0.05)
    >>> # Blobs cover a smaller volume fraction of the image
    >>> blobs = data.binary_blobs(length=256, volume_fraction=0.3)

    Notes:
    ------
    Code adapted from scikit-image
    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if isinstance(rng, int):
        rng = xp.random.RandomState(seed=rng)

    shape = tuple([length] * n_dim)
    mask = xp.zeros(shape)
    n_pts = max(int(1.0 / blob_size_fraction) ** n_dim, 1)
    points = (length * rng.rand(n_dim, n_pts)).astype(int)
    mask[tuple(indices for indices in points)] = 1

    mask = sp.ndimage.gaussian_filter(mask, sigma=0.25 * length * blob_size_fraction)
    threshold = xp.percentile(mask, 100 * (1 - volume_fraction))
    return xp.logical_not(mask < threshold)



from dexp.processing.deconvolution.admm_deconvolution import admm_deconvolution
from dexp.processing.deconvolution.blind_deconvolution import blind_deconvolution
from dexp.processing.deconvolution.inversion_deconvolution import (
    inversion_deconvolution,
)
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution



from typing import Callable

from dexp.utils import xpArray
from dexp.utils.backends import Backend


def apply_func(array: xpArray, *, func: Callable, axis: int, **kwargs) -> xpArray:

    xp = Backend.get_xp_module(array)

    results = []
    for i in range(array.shape[axis]):
        chunk = xp.take(array, i, axis=axis)
        results.append(func(chunk, **kwargs))

    return xp.stack(results, axis=axis)


def axis_wise(axis: int) -> xpArray:
    def decorator(func: Callable) -> xpArray:
        def wrapper(*args, **kwargs) -> xpArray:
            return apply_func(*args, **kwargs, func=func, axis=axis)

        return wrapper

    return decorator



from typing import Callable, Iterator, Optional, Tuple

from dexp.utils import xpArray
from dexp.utils.backends import Backend


def _get_projection_type(type: str, axis: Tuple[int]) -> Callable:
    xp = Backend.get_xp_module()
    projections = {
        "mean": lambda im: xp.mean(im, axis=axis),
        "median": lambda im: xp.median(im, axis=axis),
        "max": lambda im: xp.median(im, axis=axis),
        "min": lambda im: xp.min(im, axis=axis),
        "max-min": lambda im: xp.max(im, axis=axis) - xp.min(im, axis=axis),
    }
    try:
        return projections[type]
    except KeyError:
        raise ValueError(f"Unknown projection type: {type}")


def projection_generator(
    image: xpArray, axis_range: Optional[Tuple[int, ...]] = None, projection_type: str = "mean", nb_axis: int = 2
) -> Iterator[Tuple[int, int, xpArray]]:
    """
    Generates all nD projection of an image. Currently only supports 2D projections.

    Parameters
    ----------
    image: image to compute center of mass of.
    axis_range: Axis range to compute projections, can be: (n) for [0, n[, (m,n) for [m,n[, and (m,n,step)
        for {i*step | m<=i*step<n & i integer} projection_type: Projection type to use when in
        'projection' mode: 'mean', 'min', 'max', 'max-min'
    nb_axis: number of axis to project to, currently only supports 2D projections.

    Returns
    -------
    tuples of axis and projections: (u,v,..., projection)

    """
    if nb_axis != 2:
        # TODO: this could be fixed using itertools.combinations
        raise NotImplementedError("Not implemented for nb_axis!=2")

    ndim = image.ndim

    if axis_range is None:
        axis_range = (0, ndim, 1)

    for u in range(*axis_range):
        for v in range(*axis_range):
            if u < v:
                proj_axis = tuple(set(range(*axis_range)).difference({u, v}))
                projection = _get_projection_type(projection_type.lower(), proj_axis)
                projected_image = projection(image)
                projected_image = projected_image.astype(dtype=image.dtype, copy=False)

                yield u, v, projected_image

            elif u < 0 or v < 0 or u >= ndim or v >= ndim:
                raise ValueError(f"Axis range {axis_range} is out-of-bounds for the image of dimension: {ndim}")



from dexp.utils import xpArray
from dexp.utils.backends import Backend, CupyBackend, NumpyBackend


def nan_to_zero(array: xpArray, copy: bool = True) -> xpArray:
    """
    Replaces every nan in an array to zero. It might, or not, be able to operate in-place.
    To be safe, the returned array should always be used...

    Parameters
    ----------
    array : array to replace NaNs with zeros.
    copy : True/False to suggest whether copy or in-place behaviour should occur.

    Returns
    -------
    Array for which NaNs have been replace by zero.

    """
    # TODO: should we remove this function?
    backend = Backend.current()

    if type(backend) is NumpyBackend:
        xp = backend.get_xp_module()
        return xp.nan_to_num(array, copy=copy)
    elif type(backend) is CupyBackend:
        import cupy

        return cupy.nan_to_num(array)



from pathlib import Path
from typing import Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LogNorm
from tqdm import tqdm

from dexp.datasets import BaseDataset
from dexp.utils.backends import BestBackend


def dataset_histogram(
    dataset: BaseDataset,
    output_dir: str,
    channels: Sequence[str],
    device: int,
    minimum_count: int,
    maximum_count: Optional[float],
):

    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    for channel in channels:
        histograms = []

        data_dir = output_dir / f"{channel}_data"
        data_dir.mkdir(exist_ok=True)

        with PdfPages(output_dir / f"histograms_{channel}.pdf") as pdf:

            with BestBackend(device_id=device) as bkd:

                for i, stack in tqdm(enumerate(dataset[channel]), f"Computing Histogram of {channel}"):
                    stack = bkd.to_backend(stack)
                    hist = np.bincount(stack.flatten())

                    if minimum_count > 1:
                        (non_zero,) = np.nonzero(hist < minimum_count)
                        hist = hist[: non_zero.max()]

                    if maximum_count is not None:
                        hist = np.clip(hist, a_min=None, a_max=int(maximum_count))

                    hist = bkd.to_numpy(hist)

                    np.save(data_dir / f"{channel}_{i}.npy", hist)
                    histograms.append(hist)

                max_length = max(len(h) for h in histograms)
                max_count = max(h.max() for h in histograms)
                hist_2d = np.zeros((len(histograms), max_length), dtype=int)

                for i, hist in enumerate(histograms):
                    hist_2d[i, : len(hist)] += hist

                    plt.figure(figsize=(5, 5))
                    plt.bar(np.arange(len(hist)), hist, width=1)
                    plt.title(f"Histogram of {channel} and index {i}.")
                    plt.ylabel("Count")
                    plt.yscale("log")
                    plt.ylim(top=max_count)
                    plt.xlim(0, max_length)
                    plt.xlabel("Image intensity")
                    pdf.savefig()
                    plt.close()

                plot = plt.imshow(hist_2d, cmap="viridis", norm=LogNorm(), aspect="auto")
                plt.ylabel("Time")
                plt.xlabel("Intensity")
                plt.colorbar(plot, label="Count")
                pdf.savefig()
                plt.close()



# flake8: noqa
import numpy
from arbol import Arbol
from skimage import data
from skimage.color import rgb2gray
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from dexp.processing.denoising.noise import add_noise
from dexp.utils.backends import CupyBackend, NumpyBackend


def demo_noise_numpy():
    with NumpyBackend():
        _demo_noise()


def demo_noise_cupy():
    try:
        with CupyBackend():
            _demo_noise()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_noise(display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Arbol.enable_output = True
    Arbol.set_log_max_depth(5)

    image = data.astronaut()
    image = rgb2gray(image)

    noisy = add_noise(image)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    psnr_noisy = psnr(image.astype(noisy.dtype), noisy)
    ssim_noisy = ssim(image, noisy)
    print("         noisy   :", psnr_noisy, ssim_noisy)

    assert psnr_noisy > 12 and psnr_noisy < 13
    assert ssim_noisy > 0.28 and ssim_noisy < 0.29

    if display:
        import napari

        viewer = napari.Viewer()
        viewer.add_image(image, name="image")
        viewer.add_image(noisy, name="noisy")
        napari.run()


if __name__ == "__main__":
    demo_noise_cupy()
    demo_noise_numpy()



from typing import Dict, Sequence, Tuple

import dask.array as da
import numpy as np
from dask.array.image import imread
from numpy.typing import ArrayLike

from dexp.datasets.base_dataset import BaseDataset


class TIFDataset(BaseDataset):
    def __init__(self, path: str):
        super().__init__(dask_backed=False)
        self._channel = "stack"
        self._array = imread(path)

    def _assert_channel(self, channel: str) -> None:
        if channel != self._channel:
            raise KeyError("TIFDataset only contains a single channel named `stack`.")

    def channels(self) -> Sequence[str]:
        return ["stack"]

    def close(self) -> None:
        pass

    def shape(self, channel: str) -> Sequence[int]:
        self._assert_channel(channel)
        return self._array.shape

    def dtype(self, channel: str) -> np.dtype:
        self._assert_channel(channel)
        return self._array.dtype

    def get_metadata(self) -> Dict:
        return {}

    def append_metadata(self, metadata: Dict) -> None:
        raise NotImplementedError

    def add_channel(self, name: str, shape: Tuple[int, ...], dtype, enable_projections: bool = True, **kwargs) -> None:
        raise NotImplementedError

    @staticmethod
    def _maybe_compute(array: da.Array, wrap_with_dask: bool) -> ArrayLike:
        return array if wrap_with_dask else array.compute()

    def get_array(self, channel: str, per_z_slice: bool = False, wrap_with_dask: bool = True) -> ArrayLike:
        self._assert_channel(channel)
        return self._maybe_compute(self._array, wrap_with_dask)

    def get_stack(
        self, channel: str, time_point: int, per_z_slice: bool = False, wrap_with_dask: bool = True
    ) -> ArrayLike:
        if per_z_slice:
            raise NotImplementedError

        self._assert_channel(channel)
        return self._maybe_compute(self._array[time_point], wrap_with_dask)

    def get_projection_array(self, channel: str, axis: int, wrap_with_dask: bool = True) -> ArrayLike:
        self._assert_channel(channel)
        return self._maybe_compute(self._array.max(axis=axis), wrap_with_dask)

    def write_array(self, channel: str, array: ArrayLike) -> None:
        raise NotImplementedError

    def write_stack(self, channel: str, time_point: int, stack: ArrayLike) -> None:
        raise NotImplementedError

    def check_integrity(self, channels: Sequence[str]) -> bool:
        return True



import click
from arbol import asection

from dexp.cli.defaults import DEFAULT_CLEVEL, DEFAULT_CODEC, DEFAULT_STORE
from dexp.cli.parsing import (
    _get_output_path,
    _parse_channels,
    _parse_chunks,
    _parse_devices,
    _parse_slicing,
)
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.denoise import dataset_denoise
from dexp.datasets.zarr_dataset import ZDataset


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--channels", "-c", default=None, help="list of channels, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z) ",
)
@click.option("--tilesize", "-ts", type=int, default=320, help="Tile size for tiled computation", show_default=True)
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option("--chunks", "-chk", default=None, help="Dataset chunks dimensions, e.g. (1, 126, 512, 512).")
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="Compression codec: zstd for ’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="Forces overwrite of target", show_default=True)
@click.option(
    "--devices", "-d", type=str, default="0", help="Sets the CUDA devices id, e.g. 0,1,2 or ‘all’", show_default=True
)
def denoise(
    input_paths,
    output_path,
    channels,
    slicing,
    tilesize,
    store,
    chunks,
    codec,
    clevel,
    overwrite,
    devices,
):
    """Denoises input image using butterworth filter, parameters are estimated automatically
    using noise2self j-invariant cross-validation loss.
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)
    slicing = _parse_slicing(slicing)
    devices = _parse_devices(devices)

    if slicing is not None:
        input_dataset.set_slicing(slicing)

    output_path = _get_output_path(input_paths[0], output_path, "_denoised")
    mode = "w" + ("" if overwrite else "-")
    output_dataset = ZDataset(
        output_path,
        mode=mode,
        store=store,
        codec=codec,
        clevel=clevel,
        chunks=_parse_chunks(chunks),
        parent=input_dataset,
    )

    with asection(f"Denoising: {input_paths} to {output_path} for channels {channels}"):
        dataset_denoise(
            input_dataset,
            output_dataset,
            channels=channels,
            tilesize=tilesize,
            devices=devices,
        )

    input_dataset.close()
    output_dataset.close()



import os
from os.path import exists, join

import yaml

__config = None


def get_config():
    if __config is not None:
        return __config

    dexp_home = join(os.getenv("HOME"), ".dexp")
    os.makedirs(dexp_home, exist_ok=True)

    config_file_path = join(dexp_home, "config.yaml")
    if exists(config_file_path):
        with open(config_file_path) as stream:
            config = yaml.load(stream, Loader=yaml.FullLoader)
            return config

    else:
        config = {}
        return config


__config = get_config()



#!/usr/bin/env python

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import os
import sys
from datetime import datetime
from shutil import rmtree

from setuptools import Command, find_packages, setup

# Package meta-data.
NAME = "dexp"
DESCRIPTION = "Light-sheet Dataset EXploration and Processing"
URL = "https://github.com/royerlab/dexp"
EMAIL = "jordao.bragantini@czbiohub.org, ahmetcan.solak@czbiohub.org, bin.yang@czbiohub.org, loic.royer@czbiohub.org"
AUTHOR = "Jordao Bragantini, Ahmet Can Solak, Bin Yang, Loic A Royer"
REQUIRES_PYTHON = ">=3.7.0"

now = datetime.now()
seconds_since_midnight = (now - now.replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
minutes_since_midnight = int(seconds_since_midnight // 60)
ten_minutes_since_midnight = int(seconds_since_midnight // 600)
VERSION = datetime.today().strftime("%Y.%m.%d") + f".{minutes_since_midnight}"
print(f"Version: {VERSION}")

CUPY_VERSION = "10.1.0"

# What packages are required for this module to be executed?
REQUIRED = [
    "numpy>=1.20",
    "scipy>=1.8.0",
    "numcodecs",
    "scikit-image",
    "tifffile",
    "tqdm",
    "numexpr",
    "joblib",
    "pytest",
    "click>=8.0.0",
    "dask",
    "dask-image",
    "dask-cuda",
    "distributed",
    "zarr",
    "cachey",
    "gputil",
    "gpustat",
    "arbol",
    "colorcet",
    "PyYAML",
    "cairocffi",
    "blosc",
    "seaborn",
    "ome-zarr",
]

# What packages are optional?
EXTRAS = {
    "source": [
        f"cupy=={CUPY_VERSION}",
    ],
    "cuda114": [
        f"cupy-cuda114=={CUPY_VERSION}",
    ],
    "cuda113": [
        f"cupy-cuda113=={CUPY_VERSION}",
    ],
    "cuda112": [
        f"cupy-cuda112=={CUPY_VERSION}",
    ],
    "cuda111": [
        f"cupy-cuda111=={CUPY_VERSION}",
    ],
    "cuda110": [
        f"cupy-cuda110=={CUPY_VERSION}",
    ],
    "cuda102": [
        f"cupy-cuda102=={CUPY_VERSION}",
    ],
    "cuda101": [
        f"cupy-cuda101=={CUPY_VERSION}",
    ],
    "cuda100": [
        f"cupy-cuda100=={CUPY_VERSION}",
    ],
    "color": [
        "colorama",
        "ansicolors",
    ],
    "napari": [
        "napari[all]",
    ],
    "optional": [
        "higra",
        "pyotf",
        "tensorstore",
    ],
    "gpu": [
        "gputools",
        "cucim",
    ],
    "dev": [
        "flake8",
        "pytest",
    ],
}


# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f"\033[1m{s}\033[0m")

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        # install twine if not
        os.system("pip install -y twine --dev")

        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")

        self.status(f'Uploading version `{about["__version__"]}` of the package to PyPI via Twine…')
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    entry_points={
        "console_scripts": [
            "dexp=dexp.cli.dexp_main:cli",
            "dexp-video=dexp.cli.video_main:cli",
            "dexp-install=dexp.cli.install_main:cli",
        ]
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="BSD 3-Clause",
    license_file="LICENSE.txt",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Environment :: GPU :: NVIDIA CUDA :: 11.1",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



import click
from arbol import asection

from dexp.cli.defaults import DEFAULT_CLEVEL, DEFAULT_CODEC, DEFAULT_STORE
from dexp.cli.parsing import _get_output_path, _parse_chunks
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.fromraw import dataset_fromraw
from dexp.datasets.zarr_dataset import ZDataset


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--output_path", "-o")  # , help='output path'
@click.option("--ch-prefix", "-c", default=None, help="Prefix of channels, usually according to wavelength")
@click.option("--store", "-st", default=DEFAULT_STORE, help="Zarr store: ‘dir’, ‘ndir’, or ‘zip’", show_default=True)
@click.option("--chunks", "-chk", default=None, help="Dataset chunks dimensions, e.g. (1, 126, 512, 512).")
@click.option(
    "--codec",
    "-z",
    default=DEFAULT_CODEC,
    help="Compression codec: zstd for ’, ‘blosclz’, ‘lz4’, ‘lz4hc’, ‘zlib’ or ‘snappy’ ",
    show_default=True,
)
@click.option("--clevel", "-l", type=int, default=DEFAULT_CLEVEL, help="Compression level", show_default=True)
@click.option("--overwrite", "-w", is_flag=True, help="Forces overwrite of target", show_default=True)
@click.option(
    "--workers",
    "-wk",
    default=-4,
    help="Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ",
    show_default=True,
)  #
def fromraw(
    input_paths,
    output_path,
    ch_prefix,
    store,
    chunks,
    codec,
    clevel,
    workers,
    overwrite,
):
    """Copies a dataset, channels can be selected, cropping can be performed, compression can be changed, ..."""

    input_dataset, input_paths = glob_datasets(input_paths)
    output_path = _get_output_path(input_paths[0], output_path, "_fromraw")
    mode = "w" + ("" if overwrite else "-")
    output_dataset = ZDataset(
        output_path,
        mode=mode,
        store=store,
        codec=codec,
        clevel=clevel,
        chunks=_parse_chunks(chunks),
        parent=input_dataset,
    )

    with asection(f"Copying `fromraw`: {input_paths} to {output_path} for channels with prefix {ch_prefix}"):
        dataset_fromraw(
            input_dataset,
            output_dataset,
            ch_prefix,
            workers=workers,
        )

        input_dataset.close()



from typing import Dict, Union

import numpy

try:
    import cupy

    xpArray = Union[cupy.ndarray, numpy.ndarray]
except ImportError:
    xpArray = numpy.ndarray

from dexp.utils.backends.cupy_backend import is_cupy_available


def dict_or(lhs: Dict, rhs: Dict) -> Dict:
    """
    Can be removed once we support only >= python 3.9
    https://peps.python.org/pep-0584/
    """
    lhs = lhs.copy()
    for k, v in rhs.items():
        if k not in lhs:
            lhs[k] = v

    return lhs



from typing import Optional, Tuple

import numpy as np
import zarr

from dexp.utils.slicing import slice_from_shape


class StackIterator:
    def __init__(self, array: zarr.Array, slicing: Optional[slice]):

        self._out_shape, self._volume_slicing, self._time_points = slice_from_shape(array.shape, slicing)

        self._array = array

    @property
    def shape(self) -> Tuple[int]:
        return self._out_shape

    def __len__(self) -> int:
        return len(self._time_points)

    def __getitem__(self, index: int) -> np.ndarray:
        if isinstance(self._volume_slicing, type(...)):
            slicing = self._time_points[index]
        else:
            slicing = (self._time_points[index],) + self._volume_slicing
        return self._array[slicing]

    def time(self, index: int) -> int:
        return self._time_points[index]



from typing import Callable, Dict, Optional

import numpy as np
from numpy.typing import ArrayLike
from scipy.optimize import minimize
from toolz import curry

from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.utils import xpArray
from dexp.utils.backends import Backend


@curry
def _objfun(coefs: ArrayLike, observed: xpArray, deconved: xpArray, psf_fun: Callable) -> float:
    # TODO: deconved could be pre computed in the freq domain

    backend = Backend.current()

    psf = psf_fun(coefs)
    psf = backend.to_backend(psf)

    blurred = fft_convolve(deconved, psf, mode="reflect")
    blurred -= blurred.min()
    blurred /= blurred.max()

    return ((observed - blurred) ** 2).mean().item()


def blind_deconvolution(
    image: xpArray,
    observed_psf: xpArray,
    deconv_fun: Callable[[xpArray, xpArray], xpArray],
    microscope_params: Dict,
    n_iterations: int,
    n_zernikes: int = 15,
    display: bool = False,
) -> xpArray:
    # Local imports to
    from pyotf.phaseretrieval import PhaseRetrievalResult, retrieve_phase
    from pyotf.utils import prep_data_for_PR

    if display:
        import napari

        viewer = napari.Viewer()

    backend = Backend.current()

    image = backend.to_backend(image)

    psf = prep_data_for_PR(observed_psf)
    phase_retriv: PhaseRetrievalResult = retrieve_phase(psf, params=microscope_params)
    phase_retriv.fit_to_zernikes(n_zernikes)

    def get_psf(coefs: Optional[ArrayLike] = None) -> ArrayLike:
        if coefs is not None:
            phase_retriv.zd_result.pcoefs = coefs[: len(coefs) // 2]
            phase_retriv.zd_result.mcoefs = coefs[len(coefs) // 2 :]
        psf = phase_retriv.generate_zd_psf()
        return (psf / psf.sum()).astype(image.dtype)

    if display:
        viewer.add_image(backend.to_numpy(image), name="Input Image")
        viewer.add_image(observed_psf, name="Observed PSF")
        viewer.add_image(psf, name="Prepared PSF")

    psf = get_psf()
    if display:
        viewer.add_image(psf, name="Estimated PSF")

    coefs = np.concatenate((phase_retriv.zd_result.pcoefs, phase_retriv.zd_result.mcoefs))

    for i in range(n_iterations):

        deconv = deconv_fun(image, psf)

        objfun = _objfun(observed=image, deconved=deconv, psf_fun=get_psf)
        opt_res = minimize(objfun, x0=coefs, method="L-BFGS-B")

        old_coefs = coefs
        coefs = opt_res.x
        psf = get_psf(coefs)

        # FIXME
        print(np.square(old_coefs - coefs).mean())

        if display:
            viewer.add_image(backend.to_numpy(deconv), name=f"Deconv i={i}")
            viewer.add_image(psf, name=f"Est. PSF i={i}")

    if display:
        napari.run()

    return deconv



import functools
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import dask
import numpy
import scipy
from arbol.arbol import aprint, asection
from dask.distributed import Client
from dask_cuda import LocalCUDACluster

from dexp.datasets import BaseDataset
from dexp.optics.psf.standard_psfs import nikon16x08na, olympus20x10na
from dexp.processing.deconvolution import (
    admm_deconvolution,
    lucy_richardson_deconvolution,
)
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i
from dexp.utils.backends import Backend, BestBackend
from dexp.utils.slicing import slice_from_shape


def dataset_deconv(
    dataset: BaseDataset,
    dest_path: str,
    channels: Sequence[str],
    slicing,
    store: str = "dir",
    compression: str = "zstd",
    compression_level: int = 3,
    overwrite: bool = False,
    tilesize: Optional[Tuple[int]] = None,
    method: str = "lr",
    num_iterations: int = 16,
    max_correction: int = 16,
    power: float = 1,
    blind_spot: int = 0,
    back_projection: Optional[str] = None,
    wb_order: int = 5,
    psf_objective: str = "nikon16x08na",
    psf_na: float = 0.8,
    psf_dxy: float = 0.485,
    psf_dz: float = 2,
    psf_xy_size: int = 17,
    psf_z_size: int = 17,
    psf_show: bool = False,
    scaling: Optional[Tuple[float]] = None,
    workers: int = 1,
    workersbackend: str = "",
    devices: Optional[List[int]] = None,
    check: bool = True,
    stop_at_exception: bool = True,
):

    from dexp.datasets import ZDataset

    mode = "w" + ("" if overwrite else "-")
    dest_dataset = ZDataset(dest_path, mode, store, parent=dataset)

    # Default tile size:
    if tilesize is None:
        tilesize = 320  # very conservative

    # Scaling default value:
    if scaling is None:
        scaling = (1, 1, 1)
    sz, sy, sx = scaling
    aprint(f"Input images will be scaled by: (sz,sy,sx)={scaling}")

    # CUDA DASK cluster
    cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
    client = Client(cluster)
    aprint("Dask Client", client)

    lazy_computation = []

    for channel in dataset._selected_channels(channels):
        array = dataset.get_array(channel)

        aprint(f"Slicing with: {slicing}")
        out_shape, volume_slicing, time_points = slice_from_shape(array.shape, slicing)

        out_shape = tuple(int(round(u * v)) for u, v in zip(out_shape, (1,) + scaling))
        dtype = numpy.float16 if method == "admm" else array.dtype

        # Adds destination array channel to dataset
        dest_array = dest_dataset.add_channel(
            name=channel, shape=out_shape, dtype=dtype, codec=compression, clevel=compression_level
        )

        # This is not ideal but difficult to avoid right now:
        sxy = (sx + sy) / 2

        # PSF paraneters:
        psf_kwargs = {
            "dxy": psf_dxy / sxy,
            "dz": psf_dz / sz,
            "xy_size": int(round(psf_xy_size * sxy)),
            "z_size": int(round(psf_z_size * sz)),
        }

        aprint(f"psf_kwargs: {psf_kwargs}")

        # NA override:
        if psf_na is not None:
            aprint(f"Numerical aperture overridden to a value of: {psf_na}")
            psf_kwargs["NA"] = psf_na

        # choose psf from detection optics:
        if psf_objective == "nikon16x08na":
            psf_kernel = nikon16x08na(**psf_kwargs)
        elif psf_objective == "olympus20x10na":
            psf_kernel = olympus20x10na(**psf_kwargs)
        elif Path(psf_objective).exists():
            psf_kernel = numpy.load(psf_objective)
            if sz != 1.0 or sy != 1.0 or sx != 1.0:
                psf_kernel = scipy.ndimage.zoom(psf_kernel, zoom=(sz, sy, sx), order=1)
            psf_z_size = psf_kernel.shape[0] + 10
            psf_xy_size = max(psf_kernel.shape[1:]) + 10
        else:
            raise RuntimeError(f"Object/path {psf_objective} not found.")

        # usefull for debugging:
        if psf_show:
            import napari

            viewer = napari.Viewer(title="DEXP | viewing PSF with napari", ndisplay=3)
            viewer.add_image(psf_kernel)
            napari.run()

        margins = max(psf_xy_size, psf_z_size)

        if method == "lr":
            normalize = False
            convolve = functools.partial(fft_convolve, in_place=False, mode="reflect", internal_dtype=numpy.float32)

            def deconv(image):
                min_value = image.min()
                max_value = image.max()

                return lucy_richardson_deconvolution(
                    image=image,
                    psf=psf_kernel,
                    num_iterations=num_iterations,
                    max_correction=max_correction,
                    normalise_minmax=(min_value, max_value),
                    power=power,
                    blind_spot=blind_spot,
                    blind_spot_mode="median+uniform",
                    blind_spot_axis_exclusion=(0,),
                    wb_order=wb_order,
                    back_projection=back_projection,
                    convolve_method=convolve,
                )

        elif method == "admm":
            normalize = True

            def deconv(image):
                out = admm_deconvolution(
                    image,
                    psf=psf_kernel,
                    iterations=num_iterations,
                    derivative=2,
                )
                return out

        else:
            raise ValueError(f"Unknown deconvolution mode: {method}")

        @dask.delayed
        def process(i):
            tp = time_points[i]
            try:
                with asection(f"Deconvolving time point for time point {i}/{len(time_points)}"):
                    with asection(f"Loading channel: {channel}"):
                        tp_array = numpy.asarray(array[tp][volume_slicing])

                    with BestBackend(exclusive=True, enable_unified_memory=True):

                        if sz != 1.0 or sy != 1.0 or sx != 1.0:
                            with asection(f"Applying scaling {(sz, sy, sx)} to image."):
                                sp = Backend.get_sp_module()
                                tp_array = Backend.to_backend(tp_array)
                                tp_array = sp.ndimage.zoom(tp_array, zoom=(sz, sy, sx), order=1)
                                tp_array = Backend.to_numpy(tp_array)

                        with asection(
                            f"Deconvolving image of shape: {tp_array.shape}, with tile size: {tilesize}, "
                            + "margins: {margins} "
                        ):
                            aprint(f"Number of iterations: {num_iterations}, back_projection:{back_projection}, ")
                            tp_array = scatter_gather_i2i(
                                deconv,
                                tp_array,
                                tiles=tilesize,
                                margins=margins,
                                normalise=normalize,
                                internal_dtype=dtype,
                            )

                        with asection("Moving array from backend to numpy."):
                            tp_array = Backend.to_numpy(tp_array, dtype=dest_array.dtype, force_copy=False)

                    with asection(
                        f"Saving deconvolved stack for time point {i}, shape:{tp_array.shape}, dtype:{array.dtype}"
                    ):
                        dest_dataset.write_stack(channel=channel, time_point=i, stack_array=tp_array)

                    aprint(f"Done processing time point: {i}/{len(time_points)} .")

            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {i} !")
                import traceback

                traceback.print_exc()

                if stop_at_exception:
                    raise error

        for i in range(len(time_points)):
            lazy_computation.append(process(i))

    dask.compute(*lazy_computation)

    # Dataset info:
    aprint(dest_dataset.info())

    # Check dataset integrity:
    if check:
        dest_dataset.check_integrity()

    # close destination dataset:
    dest_dataset.close()
    client.close()



from functools import partial
from typing import Optional

import numpy as np
from scipy.signal._signaltools import _centered

from dexp.processing.morphology.utils import get_3d_image_graph
from dexp.processing.utils import apply_func
from dexp.utils import xpArray
from dexp.utils.backends import Backend


def _generic_area_filtering(
    image: xpArray,
    area_threshold: float,
    tree_type: str,
    sampling: int,
) -> xpArray:
    """Applies an area filtering a down sampled version of the input given a component tree type.
    Note: Moves the data to the CPU.

    Parameters
    ----------
    image : xpArray
        Input grayscale image.
    area_threshold : float
        Threshold of area (number of pixels) after the sampling reduction.
    tree_type : str
        Type of tree for filtering computation, "max" ("min") used for opening (closing).
    sampling : int
        Down sampling rate for component tree operation.

    Returns
    -------
    xpArray
        Area filtered result.
    """
    import higra as hg

    if tree_type.lower() == "min":
        component_tree_fun = hg.component_tree_min_tree
    elif tree_type.lower() == "max":
        component_tree_fun = hg.component_tree_max_tree
    else:
        raise NotImplementedError

    orig_shape = image.shape
    sp = Backend.get_sp_module(image)
    if sampling != 1:
        image = sp.ndimage.zoom(image, zoom=1 / sampling, order=1)

    orig_dtype = image.dtype
    image = Backend.to_numpy(image, dtype=np.float32)

    if image.ndim == 2:
        graph = hg.get_4_adjacency_graph(image.shape)
    elif image.ndim == 3:
        graph = get_3d_image_graph(image.shape)
    else:
        raise NotImplementedError

    tree, alt = component_tree_fun(graph, image)

    area = hg.attribute_area(tree)
    filtered = hg.reconstruct_leaf_data(tree, alt, area < area_threshold)

    hg.clear_auto_cache()

    filtered = Backend.to_backend(filtered, dtype=orig_dtype)
    if sampling != 1:
        filtered = sp.ndimage.zoom(filtered, zoom=sampling, order=1)

    return _centered(filtered, orig_shape)


def area_opening(
    image: xpArray,
    area_threshold: float,
    sampling: int,
    axis: Optional[int] = None,
) -> xpArray:
    """Area opening on the downsampled input.
    Note: Moves the data to the CPU.

    Parameters
    ----------
    image : xpArray
        Input grayscale image.
    area_threshold : float
        Threshold of area (number of pixels) after the sampling reduction.
    sampling : int
        Down sampling rate for component tree operation.
    axis : int
        Apply transformation of this axis individually.

    Returns
    -------
    xpArray
        Area opened result.
    """
    _area_opening = partial(_generic_area_filtering, area_threshold=area_threshold, tree_type="max", sampling=sampling)

    if axis is None:
        return _area_opening(image=image)

    return apply_func(image, func=_area_opening, axis=axis)


def area_closing(
    image: xpArray,
    area_threshold: float,
    sampling: int,
    axis: Optional[int] = None,
) -> xpArray:
    """Area closing on the downsampled input.
    Note: Moves the data to the CPU.

    Parameters
    ----------
    image : xpArray
        Input grayscale image.
    area_threshold : float
        Threshold of area (number of pixels) after the sampling reduction.
    sampling : int
        Down sampling rate for component tree operation.
    axis : int
        Apply transformation of this axis individually.

    Returns
    -------
    xpArray
        Area closed result.
    """
    _area_closing = partial(_generic_area_filtering, area_threshold=area_threshold, tree_type="min", sampling=sampling)

    if axis is None:
        return _area_closing(image=image)

    return apply_func(image, func=_area_closing, axis=axis)


def area_white_top_hat(
    image: xpArray, area_threshold: float, sampling: int, clip_negatives: bool = True, axis: Optional[int] = None
) -> xpArray:

    """Area white hot transform on the downsampled input.
    This operation returns every *bright* component smaller than the given `area_threshold`.
    The white top hat transform is I - O(I), where O(.) is the area opening operator and I the original input.
    Note: It moves the data to the CPU.

    Parameters
    ----------
    image : xpArray
        Input grayscale image.
    area_threshold : float
        Threshold of area (number of pixels) after the sampling reduction.
    sampling : int
        Down sampling rate for component tree operation.
    clip_zeros : bool
        This operation can result to negative values that can be clipped given this parameter.
    axis : int
        Apply transformation of this axis individually.

    Returns
    -------
    xpArray
        Area white top hat result.
    """
    xp = Backend.get_xp_module(image)

    opened = area_opening(image, area_threshold=area_threshold, sampling=sampling, axis=axis)

    signed_type = np.result_type(opened.dtype, image.dtype, np.byte)
    wth = image - opened.astype(signed_type)

    if not clip_negatives:
        return wth

    return xp.clip(wth, 0, None).astype(image.dtype)


def area_black_top_hat(image: xpArray, area_threshold: float, sampling: int, axis: Optional[int] = None) -> xpArray:
    """Area black hot transform on the downsampled input.
    This operation returns the complement (inverse mapping) of every *dark* component smaller than
    the given `area_threshold`.
    The black top hat transform is C(I) - I, where C(.) is the area closing operator and I the original input.
    Note: It moves the data to the CPU.

    Parameters
    ----------
    image : xpArray
        Input grayscale image.
    area_threshold : float
        Threshold of area (number of pixels) after the sampling reduction.
    sampling : int
        Down sampling rate for component tree operation.
    axis : int
        Apply transformation of this axis individually.

    Returns
    -------
    xpArray
        Area black top hat result.
    """
    closed = area_closing(image, area_threshold=area_threshold, sampling=sampling, axis=axis)
    return closed - image



import numpy as np
from arbol import aprint, asection
from joblib import Parallel, delayed
from toolz import curry

from dexp.datasets import CCDataset, ZDataset
from dexp.utils.misc import compute_num_workers


@curry
def _write(in_ds: CCDataset, out_ds: ZDataset, channel: str, indices: np.ndarray, index: int):
    aprint(f"Writing time point {index} of channel {channel}")
    stack = in_ds.get_stack(channel, indices[index])
    out_ds.write_stack(channel, index, stack)


def dataset_fromraw(
    in_dataset: CCDataset,
    out_dataset: ZDataset,
    channel_prefix: str,
    workers: int,
) -> None:

    channels = list(filter(lambda x: x.startswith(channel_prefix), in_dataset.channels()))
    aprint(f"Selected channels {channels}")

    argmin = channels[0]
    for ch in channels:
        if in_dataset.nb_timepoints(ch) < in_dataset.nb_timepoints(argmin):
            argmin = ch

    min_secs = in_dataset.time_sec(argmin)
    min_time_pts = in_dataset.nb_timepoints(argmin)

    ch_to_mask = {ch: np.ones(in_dataset.nb_timepoints(ch), dtype=bool) for ch in channels}

    aprint(f"Min. time points is {min_time_pts} at channel {argmin}.")

    for ch in channels:
        ch_secs = in_dataset.time_sec(ch)

        with asection(f"Channel {ch} has {len(ch_secs)} time points."):
            while ch_to_mask[ch].sum() > min_time_pts:
                min_shift = ch_to_mask[ch].sum() - 1
                min_cost = np.abs(ch_secs[: len(min_secs)] - min_secs).sum()  # this should be the current cost
                for i in range(min_time_pts - 1):
                    mask = ch_to_mask[ch].copy()
                    if not mask[i]:
                        continue
                    mask[i] = False
                    cost = np.abs(ch_secs[mask][: len(min_secs)] - min_secs).sum()
                    if cost < min_cost:
                        min_cost = cost
                        min_shift = i

                aprint(f"Channel {ch} min. shift at {min_shift}.")
                ch_to_mask[ch][min_shift] = False

        aprint(f"Channel {ch} min. shifts: {np.where(np.logical_not(ch_to_mask[ch]))[0]}")

    for ch in channels:
        out_dataset.add_channel(ch, in_dataset.shape(argmin), in_dataset.dtype(ch))
        write_fun = _write(in_dataset, out_dataset, ch, np.where(ch_to_mask[ch])[0])

        if workers == 1:
            for i in range(min_time_pts):
                write_fun(i)
        else:
            n_jobs = compute_num_workers(workers, min_time_pts)
            parallel = Parallel(n_jobs=n_jobs)
            parallel(delayed(write_fun)(i) for i in range(min_time_pts))

    aprint(out_dataset.info())
    out_dataset.check_integrity()
    out_dataset.close()



import os
from os.path import join
from typing import Sequence, Union

from arbol.arbol import aprint, asection
from joblib import Parallel, delayed
from tifffile import memmap

from dexp.datasets import BaseDataset
from dexp.io.io import tiff_save


def dataset_tiff(
    dataset: BaseDataset,
    dest_path: str,
    channels: Sequence[str],
    slicing,
    overwrite: bool = False,
    project: Union[int, bool] = False,
    one_file_per_first_dim: bool = False,
    clevel: int = 0,
    workers: int = 1,
    workersbackend: str = "",
    stop_at_exception: bool = True,
):

    selected_channels = dataset._selected_channels(channels)

    aprint(f"getting Dask arrays for channels {selected_channels}")
    arrays = list(dataset.get_array(channel, per_z_slice=False, wrap_with_dask=True) for channel in selected_channels)

    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        arrays = list(array[slicing] for array in arrays)
        aprint("Done slicing.")

    if workers == -1:
        workers = max(1, os.cpu_count() // abs(workers))
    aprint(f"Number of workers: {workers}")

    if one_file_per_first_dim:
        aprint(f"Saving one TIFF file for each tp (or Z if already sliced) to: {dest_path}.")

        os.makedirs(dest_path, exist_ok=True)

        def process(tp):
            try:
                with asection(f"Saving time point {tp}: "):
                    for channel, array in zip(selected_channels, arrays):
                        tiff_file_path = join(dest_path, f"file{tp}_{channel}.tiff")
                        if overwrite or not os.path.exists(tiff_file_path):
                            stack = array[tp].compute()

                            if project is not False and type(project) == int:
                                # project is the axis for projection, but here we are not considering
                                # the T dimension anymore...
                                aprint(f"Projecting along axis {project}")
                                stack = stack.max(axis=project)

                            aprint(
                                f"Writing time point: {tp} of shape: {stack.shape}, dtype:{stack.dtype} "
                                + f"as TIFF file: '{tiff_file_path}', with compression: {clevel}"
                            )
                            tiff_save(tiff_file_path, stack, compress=clevel)
                            aprint(f"Done writing time point: {tp} !")
                        else:
                            aprint(f"File for time point (or z slice): {tp} already exists.")
            except Exception as error:
                aprint(error)
                aprint(f"Error occurred while processing time point {tp} !")
                import traceback

                traceback.print_exc()

                if stop_at_exception:
                    raise error

        if workers > 1:
            Parallel(n_jobs=workers, backend=workersbackend)(
                delayed(process)(tp) for tp in range(0, arrays[0].shape[0])
            )
        else:
            for tp in range(0, arrays[0].shape[0]):
                process(tp)

    else:

        for channel, array in zip(selected_channels, arrays):
            if len(selected_channels) > 1:
                tiff_file_path = f"{dest_path}_{channel}.tiff"
            else:
                tiff_file_path = f"{dest_path}.tiff"

            if not overwrite and os.path.exists(tiff_file_path):
                aprint(f"File {tiff_file_path} already exists! Set option -w to overwrite.")
                return

            with asection(
                f"Saving array ({array.shape}, {array.dtype}) for channel {channel} into "
                + f"TIFF file at: {tiff_file_path}:"
            ):

                shape = array.shape

                if project is not False and type(project) == int:
                    shape = list(shape)
                    shape.pop(1 + project)
                    shape = tuple(shape)

                memmap_image = memmap(tiff_file_path, shape=shape, dtype=array.dtype, bigtiff=True, imagej=True)

                def process(tp):
                    aprint(f"Processing time point {tp}")
                    stack = array[tp].compute()

                    if project is not False and type(project) == int:
                        # project is the axis for projection, but here we are not considering the T dimension anymore...
                        aprint(f"Projecting along axis {project}")
                        stack = stack.max(axis=project)

                    memmap_image[tp] = stack

                if workers > 1:
                    Parallel(n_jobs=workers, backend=workersbackend)(
                        delayed(process)(tp) for tp in range(0, array.shape[0])
                    )
                else:
                    for tp in range(0, array.shape[0]):
                        process(tp)

                memmap_image.flush()
                del memmap_image



import os
import subprocess

from arbol import aprint, asection


def robocopy(
    source_folder: str,
    dest_folder: str,
    move_files: bool = False,
    nb_threads: int = 8,
    large_files: bool = False,
    wait_to_finish: bool = True,
):
    """
    Start a generic robocopy job on Windows to copy files from one folder to another

    Args
    ---------

    Returns
    ---------
    copyProcess (Popen Object) :
    > this is the Popen object that's running the copy process. This can be
    > used to check or kill the process if necessary.
    """

    with asection(
        f"Starting a Windows Robocopy job: copying all files and folders from {source_folder} "
        + f"to {dest_folder} with {nb_threads} threads."
    ):
        # checks for network file paths in source and destination paths
        if "\\" in source_folder[:1]:
            source_folder = "\\" + source_folder

        if "\\" in dest_folder[:1]:
            dest_folder = "\\" + dest_folder

        # replaces slashes:
        source = (source_folder).replace("/", "\\")
        dest = (dest_folder).replace("/", "\\")

        # logfile:
        log_file = "robocopy_log.txt"
        if os.path.exists(log_file):
            os.remove(log_file)

        # format the command list for Popen
        copyCommand = [
            "ROBOCOPY",
            source,
            dest,
            "/e",
            "/R:10",
            "/W:5",
            "/TBD",
            "/NP",
            # '/V',
            "/eta",
            "/tee",
            f"/MT:{nb_threads}",
            f"/log:{log_file}",
        ]

        if move_files:
            copyCommand.append("/move")

        if large_files:
            copyCommand.append("/j")

        # use subprocess to start a copy command
        aprint(f"Robocopy command: {copyCommand}")
        copy_process = subprocess.Popen(copyCommand)

        if wait_to_finish:
            copy_process.wait()

    return copy_process



import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _parse_channels
from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--channels", "-c", default=None, help="List of channels, all channels when ommited.")
def check(input_paths, channels):
    """Checks the integrity of a dataset."""

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)

    with asection(f"checking integrity of datasets {input_paths}, channels: {channels}"):
        result = input_dataset.check_integrity(channels)
        input_dataset.close()

        if not result:
            aprint("!!! PROBLEM DETECTED, CORRUPTION LIKELY !!!")
            return

    aprint("No problem detected.")
    aprint("Done!")



import click
from arbol.arbol import aprint, asection

from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.serve import dataset_serve


@click.command()
@click.argument("input_paths", nargs=-1)  # ,  help='input path'
@click.option("--host", "-h", type=str, default="0.0.0.0", help="Host to serve from", show_default=True)
@click.option("--port", "-p", type=int, default=8000, help="Port to serve from", show_default=True)
def serve(input_paths, host, port):
    """Serves dataset across network."""

    input_dataset, input_paths = glob_datasets(input_paths)

    with asection(f"Serving dataset(s): {input_paths}"):
        dataset_serve(input_dataset, host=host, port=port)

        input_dataset.close()
        aprint("Done!")



import click
from arbol.arbol import aprint, asection

from dexp.cli.defaults import DEFAULT_WORKERS_BACKEND
from dexp.cli.parsing import (
    _get_output_path,
    _parse_channels,
    _parse_devices,
    _parse_slicing,
)
from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.projrender import dataset_projection_rendering


@click.command()
@click.argument("input_paths", nargs=-1)
@click.option(
    "--output_path",
    "-o",
    type=str,
    default=None,
    help="Output folder to store rendered PNGs. Default is: frames_<channel_name>",
)
@click.option("--channels", "-c", type=str, default=None, help="list of channels to color, all channels when ommited.")
@click.option(
    "--slicing",
    "-s",
    type=str,
    default=None,
    help="dataset slice (TZYX), e.g. [0:5] (first five stacks) [:,0:100] (cropping in z).",
)
@click.option(
    "--overwrite", "-w", is_flag=True, help="to force overwrite of target", show_default=True
)  # , help='dataset slice'
@click.option(
    "--axis", "-ax", type=int, default=0, help="Sets the projection axis: 0->Z, 1->Y, 2->X ", show_default=True
)  # , help='dataset slice'
@click.option(
    "--dir",
    "-di",
    type=int,
    default=-1,
    help="Sets the projection direction: -1 -> top to bottom, +1 -> bottom to top.",
    show_default=True,
)  # , help='dataset slice'
@click.option(
    "--mode",
    "-m",
    type=str,
    default="colormax",
    help="Sets the projection mode: ‘max’: classic max projection, ‘colormax’: color max projection, i.e. color codes for depth, ‘maxcolor’ same as colormax but first does depth-coding by color and then max projects (acheives some level of transparency). ",
    show_default=True,
)
@click.option(
    "--clim",
    "-cl",
    type=str,
    default=None,
    help="Sets the contrast limits, i.e. -cl 0,1000 sets the contrast limits to [0,1000]",
)
@click.option(
    "--attenuation",
    "-at",
    type=float,
    default=0.1,
    help="Sets the projection attenuation coefficient, should be within [0, 1] ideally close to 0. Larger values mean more attenuation.",
    show_default=True,
)  # , help='dataset slice'
@click.option(
    "--gamma",
    "-g",
    type=float,
    default=1.0,
    help="Sets the gamma coefficient pre-applied to the raw voxel values (before projection or any subsequent processing).",
    show_default=True,
)  # , help='dataset slice'
@click.option(
    "--dlim",
    "-dl",
    type=str,
    default=None,
    help="Sets the depth limits. Depth limits. For example, a value of (0.1, 0.7) means that the colormap start at a normalised depth of 0.1, and ends at a normalised depth of 0.7, other values are clipped. Only used for colormax mode.",
    show_default=True,
)  # , help='dataset slice'
@click.option(
    "--colormap",
    "-cm",
    type=str,
    default=None,
    help="sets colormap, e.g. viridis, gray, magma, plasma, inferno. Use a rainbow colormap such as turbo, bmy, or rainbow (recommended) for color-coded depth modes. ",
    show_default=True,
)
@click.option(
    "--rgbgamma",
    "-cg",
    type=float,
    default=1.0,
    help="Gamma correction applied to the resulting RGB image. Usefull to brighten image",
    show_default=True,
)
@click.option(
    "--transparency",
    "-t",
    is_flag=True,
    help="Enables transparency output when possible. Good for rendering on white (e.g. on paper).",
    show_default=True,
)
@click.option(
    "--legendsize",
    "-lsi",
    type=float,
    default=1.0,
    help="Multiplicative factor to control size of legend. If 0, no legend is generated.",
    show_default=True,
)
@click.option(
    "--legendscale",
    "-lsc",
    type=float,
    default=1.0,
    help="Float that gives the scale in some unit of each voxel (along the projection direction). Only in color projection modes.",
    show_default=True,
)
@click.option(
    "--legendtitle",
    "-lt",
    type=str,
    default="color-coded depth (voxels)",
    help="Title for the color-coded depth legend.",
    show_default=True,
)
@click.option(
    "--legendtitlecolor",
    "-ltc",
    type=str,
    default="1,1,1,1",
    help="Legend title color as a tuple of normalised floats: R, G, B, A  (values between 0 and 1).",
    show_default=True,
)
@click.option(
    "--legendposition",
    "-lp",
    type=str,
    default="bottom_left",
    help="Position of the legend in pixels in natural order: x,y. Can also be a string: bottom_left, bottom_right, top_left, or top_right.",
    show_default=True,
)
@click.option(
    "--legendalpha",
    "-la",
    type=float,
    default=1,
    help="Transparency for legend (1 means opaque, 0 means completely transparent)",
    show_default=True,
)
@click.option("--step", "-sp", type=int, default=1, help="Process every ‘step’ frames.", show_default=True)
@click.option(
    "--workers",
    "-k",
    type=int,
    default=-1,
    help="Number of worker threads to spawn, if -1 then num workers = num devices",
    show_default=True,
)  #
@click.option(
    "--workersbackend",
    "-wkb",
    type=str,
    default=DEFAULT_WORKERS_BACKEND,
    help="What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ",
    show_default=True,
)  #
@click.option(
    "--devices", "-d", type=str, default="0", help="Sets the CUDA devices id, e.g. 0,1,2 or ‘all’", show_default=True
)  #
def projrender(
    input_paths,
    output_path,
    channels,
    slicing,
    overwrite,
    axis,
    dir,
    mode,
    clim,
    attenuation,
    gamma,
    dlim,
    colormap,
    rgbgamma,
    transparency,
    legendsize,
    legendscale,
    legendtitle,
    legendtitlecolor,
    legendposition,
    legendalpha,
    step,
    workers,
    workersbackend,
    devices,
    stop_at_exception=True,
):
    """Renders datatset using 2D projections."""

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)
    slicing = _parse_slicing(slicing)
    devices = _parse_devices(devices)

    output_path = _get_output_path(input_paths[0], output_path, f"_{mode}_projection")

    dlim = None if dlim is None else tuple(float(strvalue) for strvalue in dlim.split(","))
    legendtitlecolor = tuple(float(v) for v in legendtitlecolor.split(","))

    if "," in legendposition:
        legendposition = tuple(float(strvalue) for strvalue in legendposition.split(","))

    with asection(
        f"Projection rendering of: {input_paths} to {output_path} for channels: {channels}, slicing: {slicing} "
    ):
        dataset_projection_rendering(
            input_dataset,
            output_path,
            channels,
            slicing,
            overwrite,
            axis,
            dir,
            mode,
            clim,
            attenuation,
            gamma,
            dlim,
            colormap,
            rgbgamma,
            transparency,
            legendsize,
            legendscale,
            legendtitle,
            legendtitlecolor,
            legendposition,
            legendalpha,
            step,
            workers,
            workersbackend,
            devices,
            stop_at_exception,
        )

        input_dataset.close()
        aprint("Done!")
